"""
The spherical harmonic classes store the respective definitions and basis
functions, as well as their derivative and inverse. Please refer to the
:doc:`/theory/spherical_harmonic_definition` page for more specific information
on the mathematical definitions of spherical harmonics.
"""
import numpy as np
import pyfar as pf
import spharpy
from abc import ABC, abstractmethod


class _SphericalHarmonicBase(ABC):
    """Base class defining properties to parametrize spherical harmonics.

    This base class serves as a base for all classes requiring a definition of
    the spherical harmonics without explicitly setting a spherical harmonic
    order. This class is intended for cases where the spherical harmonic order
    is implemented as a read-only property in child classes, for example when
    the order is implicitly defined by other parameters or inferred from data.

    Attributes
    ----------
    basis_type : str
        Type of spherical harmonic basis, either ``'real'`` or ``'complex'``.
        The default is ``'real'``.
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'maxN'``,
        ``'SN3D'``, or ``'SNM'``. The default is ``'N3D'``. Note that
        ``'maxN'`` is only supported up to 3rd order.
    channel_convention : str, optional
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        The default is ``'ACN'``. Note that ``'FuMa'`` is only supported up to
        3rd order.
    condon_shortley : bool, str, optional
        Condon-Shortley phase term. If ``True``, Condon-Shortley is included,
        if ``False`` it is not included. The default is ``'auto'``, which
        corresponds to ``True`` for type ``complex`` and ``False`` for type
        ``real``.
    """

    def __init__(
            self,
            basis_type="real",
            normalization="N3D",
            channel_convention="ACN",
            condon_shortley="auto",
        ):
        self._basis_type = None
        self._channel_convention = None
        self._condon_shortley = None
        self._normalization = None

        # basis_type needs to be initialized first, since the default for the
        # Condon-Shortley phase depends on the basis type
        self.basis_type = basis_type
        self.condon_shortley = condon_shortley
        # n_max needs to be initialized before channel_convention and
        # normalization, since both have restrictions on n_max
        self.channel_convention = channel_convention
        self.normalization = normalization

    @property
    @abstractmethod
    def n_max(self):
        """Get or set the spherical harmonic order."""

    @property
    def condon_shortley(self):
        """Get or set the Condon-Shortley phase term."""
        return self._condon_shortley

    @condon_shortley.setter
    def condon_shortley(self, value):
        """Get or set the Condon-Shortley phase term."""
        if isinstance(value, str):
            if value != 'auto':
                raise ValueError(
                    "condon_shortley must be a bool or the string 'auto'")

            value = self.basis_type == "complex"
        elif not isinstance(value, bool):
            raise ValueError(
                "condon_shortley must be a bool or the string 'auto'")

        if self._condon_shortley != value:
            self._condon_shortley = value
            self._on_property_change()

    @property
    def basis_type(self):
        """Get or set the spherical harmonic basis type."""
        return self._basis_type

    @basis_type.setter
    def basis_type(self, value):
        """Get or set the spherical harmonic basis type."""
        if value not in ["complex", "real"]:
            raise ValueError(
                "Invalid basis type, only 'complex' and 'real' are supported")

        if self._basis_type != value:
            self._basis_type = value
            self._on_property_change()

    @property
    def channel_convention(self):
        """Get or set the channel ordering convention."""
        return self._channel_convention

    @channel_convention.setter
    def channel_convention(self, value):
        """Get or set the channel order convention."""
        if value not in ["ACN", "FuMa"]:
            raise ValueError("Invalid channel convention, "
                             "currently only 'ACN' "
                             "and 'FuMa' are supported")

        if value == "FuMa" and self.n_max > 3:
            raise ValueError(
                "n_max > 3 is not allowed with 'FuMa' channel convention")

        if self._channel_convention != value:
            self._channel_convention = value
            self._on_property_change()

    @property
    def normalization(self):
        """Get or set the normalization convention."""
        return self._normalization

    @normalization.setter
    def normalization(self, value):
        """Get or set the normalization convention."""
        if value not in ["N3D", "NM", "maxN", "SN3D", "SNM"]:
            raise ValueError(
                "Invalid normalization, "
                "currently only 'N3D', 'NM', 'maxN', 'SN3D', 'SNM' are "
                "supported",
            )

        if value == "maxN" and self.n_max > 3:
            raise ValueError(
                "n_max > 3 is not allowed with 'maxN' normalization")

        if self._normalization != value:
            self._normalization = value
            self._on_property_change()

    def _on_property_change(self):  # noqa: B027
        """Method called when a class property changes.
        This method can be overridden in child classes to re-compute dependent
        properties.
        """
        pass


class SphericalHarmonicDefinition(_SphericalHarmonicBase):
    """Class storing the (discrete) definition of spherical harmonics.

    This class can serve as a container to create related objects, e.g.,
    spherical harmonic basis matrices for given sampling points, transforms,
    or other spherical harmonic related data and computations.

    Parameters
    ----------
    n_max : int, optional
        Maximum spherical harmonic order. Must be an integer greater or equal
        to 0. The default is ``0``.
    basis_type : str
        Type of spherical harmonic basis, either ``'complex'`` or
        ``'real'``. The default is ``'real'``.
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'SN3D'``,
        ``'SNM'``, or ``'maxN'``. ``'maxN'`` is only supported up to 3rd order.
        The default is ``'N3D'``.
    channel_convention : str, optional
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        ``'FuMa'`` is only supported up to 3rd order.
        The default is ``'ACN'``.
    condon_shortley : bool, str, optional
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` it is not
        included. The default is ``'auto'``, which corresponds to
        ``True`` for complex `basis_type` and ``False`` for real `basis_type`.
    """

    def __init__(
            self,
            n_max=0,
            basis_type="real",
            normalization="N3D",
            channel_convention="ACN",
            condon_shortley="auto",
        ):
        self._n_max = 0
        super().__init__(
            basis_type=basis_type,
            normalization=normalization,
            channel_convention=channel_convention,
            condon_shortley=condon_shortley,
        )
        self.n_max = n_max

    @property
    def n_max(self):
        """Get or set the spherical harmonic order."""
        return self._n_max

    @n_max.setter
    def n_max(self, value : int):
        """Get or set the spherical harmonic order."""
        if value < 0 or value % 1 != 0:
            raise ValueError("n_max must be a positive integer")
        if self.channel_convention == "FuMa" and value > 3:
            raise ValueError(
                "n_max > 3 is not allowed with 'FuMa' channel convention")
        if self.normalization == "maxN" and value > 3:
            raise ValueError(
                "n_max > 3 is not allowed with 'maxN' normalization")

        if self._n_max != value:
            self._n_max = value
            self._on_property_change()


class SphericalHarmonics(SphericalHarmonicDefinition):
    r"""
    Compute spherical harmonic basis matrices, their inverses, and gradients.

    The the basis functions that are used to build the matrices can be
    configured with the parameters described below.

    Parameters
    ----------
    n_max : int
        Maximum spherical harmonic order. Must be an integer greater or equal
        to 0.
    coordinates : :py:class:`~pyfar.Coordinates`, :py:class:`SamplingSphere`
        The sampling points for which the matrices are computed.
    basis_type : str, optional
        Type of spherical harmonic basis, either ``'complex'`` or
        ``'real'``. The default is ``'real'``.
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'SN3D'``,
        ``'SNM'``, or ``'maxN'``. ``'maxN'`` is only supported up to 3rd order.
        The default is ``'N3D'``.
    channel_convention : str, optional
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        ``'FuMa'`` is only supported up to 3rd order.
        The default is ``'ACN'``.
    inverse_method : str, ``None``, optional
        Method for computing the inverse transform:

        - ``'quadrature'``: compute the inverse via numerical quadrature. Note
          that this requires `coordinates` to be a :py:class:`SamplingSphere`
          object with weights summing to :math:`4\pi`.
        - ``'pseudo_inverse'``: compute the inverse via the pseudo-inverse.
          Note that this requires `coordinates` to be a
          :py:class:`SamplingSphere` object.
        - ``None``: denotes that the inverse is not defined and cannot be
          computed.
        - ``'auto'``: If coordinates are a :py:class:`SamplingSphere`, use
          ``'quadrature'`` when applicable and ``'pseudo_inverse'`` otherwise.
          If coordinates are a :py:class:`~pyfar.Coordinates` object ``None``
          is used.

        The default is ``'auto'``.

    condon_shortley : bool or str, optional
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` it is not
        included. The default is ``'auto'``, which corresponds to
        ``True`` for complex `basis_type` and ``False`` for real `basis_type`.
    """

    def __init__(
        self,
        n_max,
        coordinates,
        basis_type="real",
        normalization="N3D",
        channel_convention="ACN",
        inverse_method="auto",
        condon_shortley="auto",
    ):
        super().__init__(
            n_max=n_max,
            basis_type=basis_type,
            normalization=normalization,
            channel_convention=channel_convention,
            condon_shortley=condon_shortley,
        )

        # initialize private attributes
        self._coordinates = pf.Coordinates()
        self._inverse_method = None
        self._reset_compute_attributes()

        self.coordinates = coordinates
        self.inverse_method = inverse_method

    @classmethod
    def from_definition(cls, definition, coordinates, inverse_method="auto"):
        r"""
        Create SphericalHarmonics instance from SphericalHarmonicDefinition.

        Parameters
        ----------
        definition : SphericalHarmonicDefinition
            The spherical harmonic definition.
        coordinates : :py:class:`~pyfar.Coordinates`, :py:class:`SamplingSphere`
            The sampling points for which the matrices are computed.
        inverse_method : str, ``None``, optional
            Method for computing the inverse transform:

            - ``'quadrature'``: compute the inverse via numerical quadrature.
              Note that this requires `coordinates` to be a
              :py:class:`SamplingSphere` object with weights summing to
              :math:`4\pi`.
            - ``'pseudo_inverse'``: compute the inverse via the pseudo-inverse.
              Note that this requires `coordinates` to be a
              :py:class:`SamplingSphere` object.
            - ``None``: denotes that the inverse is not defined and cannot be
              computed.
              - ``'auto'``: If coordinates are a :py:class:`SamplingSphere`,
              use ``'quadrature'`` when applicable and ``'pseudo_inverse'``
              otherwise. If coordinates are a :py:class:`~pyfar.Coordinates`
              object ``None`` is used.

            The default is ``'auto'``.

        Returns
        -------
        SphericalHarmonics
            A new SphericalHarmonics instance initialized with the parameters
            from the definition object.
        """  # noqa: E501

        if type(definition) is not SphericalHarmonicDefinition:
            raise TypeError('definition must be a SphericalHarmonicDefinition')

        return cls(definition.n_max, coordinates, definition.basis_type,
                   definition.normalization, definition.channel_convention,
                   inverse_method, definition.condon_shortley)

    @property
    def coordinates(self):
        """Get or set the coordinates."""
        return self._coordinates

    @coordinates.setter
    def coordinates(self, value):
        """Get or set the coordinates."""
        if not isinstance(value, pf.Coordinates):
            raise TypeError("coordinates must be a pyfar.Coordinates "
                            "object or spharpy.SamplingSphere object")
        if value.cdim != 1:
            raise ValueError("Coordinates must be a 1D array")
        if value.csize == 0:
            raise ValueError("Coordinates cannot be empty")
        if value != self._coordinates:
            self._reset_compute_attributes()
            self._coordinates = value

    @property
    def inverse_method(self):
        """Get or set the inverse transform method."""
        return self._inverse_method

    @inverse_method.setter
    def inverse_method(self, value):
        """Get or set the inverse transform method."""
        if value not in ["pseudo_inverse", "quadrature", "auto", None]:
            raise ValueError(
                "Invalid inverse_method. Allowed: 'pseudo_inverse', "
                "'quadrature', or 'auto'.")

        if value == self._inverse_method:
            return

        if value is None:
            self._inverse_method = value
            return

        if type(self.coordinates) is pf.Coordinates:
            if value == "auto":
                self.inverse_method = None
                return

            raise ValueError(
                "The inverse method can only be set if the coordinates "
                "are a provided as SamplingSphere.")

        elif type(self.coordinates) is spharpy.SamplingSphere:
            if value == "auto":

                value = (
                    "quadrature"
                    if self.coordinates.quadrature
                    else "pseudo_inverse"
                )

            elif value == "quadrature" and not self.coordinates.quadrature:
                raise ValueError(
                    "'quadrature' requires `coordinates` to be a '"
                    "SamplingSphere and coordinates.quadrature to be "
                    "True.")

        self._reset_compute_attributes()
        self._inverse_method = value

    @property
    def basis(self):
        """Get the spherical harmonic basis matrix."""
        if self._basis is None:
            self._compute_basis()
        return self._basis

    @property
    def basis_gradient_theta(self):
        """
        Get the gradient of the basis matrix with respect to the
        `colatitude`/`elevation` angle.

        See the :py:mod:`~pyfar.classes.coordinates` documentation for the
        definition of the angles.
        """
        if self._basis_gradient_theta is None:
            self._compute_basis_gradient()
        return self._basis_gradient_theta

    @property
    def basis_gradient_phi(self):
        """
        Get the gradient of the basis matrix with respect to the
        `azimuth`/`polar` angle.

        See the :py:mod:`~pyfar.classes.coordinates` documentation for the
        definition of the angles.
        """
        if self._basis_gradient_phi is None:
            self._compute_basis_gradient()
        return self._basis_gradient_phi

    def _compute_basis(self):
        """
        Compute the basis matrix for the SphericalHarmonics class.
        """
        if self.basis_type == "complex":
            function = spharpy.spherical.spherical_harmonic_basis
        elif self.basis_type == "real":
            function = spharpy.spherical.spherical_harmonic_basis_real
        else:
            raise ValueError(
                "Invalid basis type, should be either 'complex' or 'real'")
        self._basis = function(
            n_max=self.n_max,
            coordinates=self.coordinates,
            normalization=self.normalization,
            channel_convention=self.channel_convention,
            condon_shortley=self.condon_shortley,
        )

    def _compute_basis_gradient(self):
        """
        Compute the gradient of the basis matrix for the SphericalHarmonics
        class.
        """
        if any((self.normalization in ["maxN", "SN3D"],
                self.channel_convention == "fuma")):
            raise ValueError(
            f"Gradient computation not supported for normalization "
            f"'{self.normalization}' and "
            f"channel convention '{self.channel_convention}'.")

        if self.basis_type == "complex":
            function = spharpy.spherical.spherical_harmonic_basis_gradient
        elif self.basis_type == "real":
           function = spharpy.spherical.spherical_harmonic_basis_gradient_real
        else:
            raise ValueError(
                "Invalid basis type, should be either 'complex' or 'real'")
        self._basis_gradient_theta, self._basis_gradient_phi = function(
            self.n_max, self.coordinates)

    @property
    def basis_inv(self):
        """Get the inverse basis matrix."""
        if self._inverse_method is None:
            raise ValueError("The inverse method is not defined.")

        if self._basis is None:
            self._compute_basis()
        if self._basis_inv is None:
            self._compute_inverse()
        return self._basis_inv

    def _compute_inverse(self):
        """
        Compute the inverse basis matrix for the SphericalHarmonics class.
        """
        if self._basis is None:
            self._compute_basis()
        _inv_flag = self.inverse_method
        if _inv_flag == "pseudo_inverse":
            self._basis_inv = np.linalg.pinv(self._basis)
        elif _inv_flag == "quadrature":
            self._basis_inv = np.einsum(
                'ij,i->ji', np.conj(self._basis), self.coordinates.weights)

    def _reset_compute_attributes(self):
        """Reset the computed attributes for the SphericalHarmonics class in
        case of changes in the parameters.
        """
        self._basis = None
        self._basis_gradient_theta = None
        self._basis_gradient_phi = None
        self._basis_inv = None

    def _on_property_change(self):
        """Reset computed attributes on property changes."""
        self._reset_compute_attributes()



from spharpy import SamplingSphere
from pyfar import Coordinates
import numpy as np
import numpy.testing as npt
import pytest
from spharpy.samplings import gaussian


def test_sampling_sphere_init():
    sampling = SamplingSphere()
    assert isinstance(sampling, SamplingSphere)


def test_sampling_sphere_init_value():
    sampling = SamplingSphere(1, 0, 0, 0)
    assert isinstance(sampling, SamplingSphere)


def test_sampling_sphere_from_coordinates():
    """Test converting Coordinates to SamplingSphere."""

    coordinates = Coordinates([1, 0, -1, 0], [0, 1, 0, -1], 0,
                              weights=[np.pi, np.pi, np.pi, np.pi])
    sampling_sphere = SamplingSphere.from_coordinates(coordinates)

    # check data in sampling_sphere
    assert type(sampling_sphere) is SamplingSphere
    npt.assert_equal(sampling_sphere.cartesian, coordinates.cartesian)
    npt.assert_equal(sampling_sphere.weights, coordinates.weights)
    assert sampling_sphere.n_max is None
    assert sampling_sphere.radius_tolerance == 1e-6
    assert sampling_sphere.quadrature_tolerance == 1e-10
    assert sampling_sphere.comment == coordinates.comment

    # make sure mutable arrays are copied from coordinates
    coordinates.weights = None
    assert sampling_sphere.weights is not None

    coordinates.x = [1, 1, 1, 1]
    npt.assert_equal(sampling_sphere.x, [1, 0, -1, 0])

    coordinates.y = [1, 1, 1, 1]
    npt.assert_equal(sampling_sphere.y, [0, 1, 0, -1])

    coordinates.z = [1, 1, 1, 1]
    npt.assert_equal(sampling_sphere.z, [0, 0, 0, 0])


def sampling_cube():
    """Helper function returning a cube sampling."""
    x = [1, -1, 0, 0, 0, 0]
    y = [0, 0, 1, -1, 0, 0]
    z = [0, 0, 0, 0, 1, -1]

    return x, y, z


def test_getter_n_max():
    x, y, z = sampling_cube()
    n_max = 1
    sampling = SamplingSphere(x, y, z, n_max)

    assert sampling.n_max == n_max


def test_setter_n_max():
    x, y, z = sampling_cube()
    n_max = 1
    sampling = SamplingSphere(x, y, z, 0)

    sampling.n_max = n_max
    assert sampling._n_max == n_max


def test_error_multiple_radius_initialization():
    """
    Test if entering points with multiple radii during initialization raises
    an error.
    """

    match = '1 m, which exceeds the tolerance of 1e-06 m'
    with pytest.raises(ValueError, match=match):
        SamplingSphere([1, 0], 0, 0)


def test_error_multiple_radius_setter():
    """
    Test if entering points with multiple radii after initialization raises
    an error.
    """

    sampling_sphere = SamplingSphere([1, 1], 0, 0)

    match = '1 m, which exceeds the tolerance of 1e-06 m'
    with pytest.raises(ValueError, match=match):
        sampling_sphere.x = [1, 0]


@pytest.mark.parametrize('sampling', [
    SamplingSphere([1, 1], 0, 0),
    SamplingSphere.from_cartesian([1, 1], 0, 0),
    SamplingSphere.from_spherical_elevation([0, 0], 0, 1),
    SamplingSphere.from_spherical_colatitude([0, 0], 0, 1),
    SamplingSphere.from_spherical_side([0, 0], 0, 1),
    SamplingSphere.from_spherical_front([0, 0], 0, 1),
    SamplingSphere.from_cylindrical([0, 0], 0, 1),
])
def test_radius_tolerance(sampling):
    """
    Test getter and setter for radius tolerance and the related error message.
    """
    tolerance = 1e-3

    # test default value
    assert sampling.radius_tolerance == 1e-6
    # change tolerance
    sampling.radius_tolerance = tolerance
    assert sampling.radius_tolerance == tolerance

    with pytest.raises(ValueError, match=f'{tolerance:.3g}'):
        sampling.x = [0, 1]


def test_radius_tolerance_error():
    """
    Test if setting the radius tolerance too strict raises an error for
    existing data.
    """
    sampling = SamplingSphere([1, 1.1], 0, 0, radius_tolerance=.2)

    with pytest.raises(ValueError, match='the tolerance of 0.01'):
        sampling.radius_tolerance = .01


@pytest.mark.parametrize('tolerance', [
    None, [0, 1], np.array([0, 1]), -.1,
])
def test_radius_tolerance_input(tolerance):
    """Test if passing wrong values raises the expected error."""

    match = 'The radius tolerance must be a number greater than zero'
    with pytest.raises(ValueError, match=match):
        SamplingSphere([1, 1], 0, 0, radius_tolerance=tolerance)


def test_weights_getter():
    x, y, z = sampling_cube()
    n_max = 1
    weights = np.ones(6)*4*np.pi/6
    sampling = SamplingSphere(x, y, z, n_max, weights=weights)

    np.testing.assert_allclose(sampling.weights, weights)


def test_setting_weights():
    x, y, z = sampling_cube()
    n_max = 1
    sampling = SamplingSphere(x, y, z, n_max)

    weights = np.ones(6)*4*np.pi/6
    sampling.weights = weights

    np.testing.assert_array_equal(sampling._weights, weights)
    assert sampling._weights.shape == (6,)


def test_setting_weights_invalid():
    x, y, z = sampling_cube()
    n_max = 1
    sampling = SamplingSphere(x, y, z, n_max)

    message = r"The sum of the weights must be equal to 4\*pi."
    with pytest.raises(ValueError, match=message):
        sampling.weights = np.ones(6)/6

    with pytest.raises(ValueError, match=message):
        sampling._set_weights(np.ones(6)/6)

    message = "All weights must be positive."
    weights_invalid = np.ones(6)*4*np.pi/6
    weights_invalid[0] = np.nan
    with pytest.raises(ValueError, match=message):
        sampling.weights = weights_invalid

    weights_invalid = np.ones(6)*4*np.pi/6 * -1
    weights_invalid[0] = -1
    with pytest.raises(ValueError, match=message):
        sampling.weights = weights_invalid


def test_quadrature_getter_changing_weights():
    # create a quadrature grid (gaussian)
    sampling = gaussian(n_max=3)
    # check if quadrature is set properly
    assert sampling.quadrature
    # update weights such that quadrature requirement is not valid anymore
    weights = 4 * np.pi * np.ones(sampling.cshape) / sampling.cshape
    sampling.weights = weights
    assert not sampling.quadrature


def test_quadrature_getter_changing_points():
    # create a quadrature grid (gaussian)
    sampling = gaussian(n_max=3)
    # check if quadrature is set properly
    assert sampling.quadrature
    # update points such that quadrature is not valid anymore
    rng = np.random.default_rng()
    sampling.spherical_colatitude = np.concatenate([
        rng.random((sampling.csize, 1)),
        np.ones((sampling.csize, 1)),
        np.ones((sampling.csize, 1))], axis=1)
    assert not sampling.quadrature


def test_repr():
    """Test representation string."""

    sampling = SamplingSphere([1, -1], 0, 0)
    repr_str = sampling.__repr__()
    assert repr_str == 'SamplingSphere: n_max=None, cshape=(2,)'

    sampling = SamplingSphere([1, -1], 0, 0, n_max=0)
    repr_str = sampling.__repr__()
    assert repr_str == 'SamplingSphere: n_max=0, cshape=(2,)'



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
import urllib3
import shutil
import numpy as np
sys.path.insert(0, os.path.abspath('..'))
import spharpy  # noqa: E402

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.mathjax',
    'sphinx.ext.intersphinx',
    'autodocsumm',
    'sphinx_design',
    'sphinx_favicon',
    'sphinx_reredirects',
    'sphinx_mdinclude',
    'sphinx_copybutton',
]

# show tocs for classes and functions of modules using the autodocsumm
# package
autodoc_default_options = {'autosummary': True}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'spharpy'
copyright = "2020 - 2023, Marco Berzborn; 2023, The pyfar developers"
author = "The pyfar developers"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = spharpy.__version__
# The full version, including alpha/beta/rc tags.
release = spharpy.__version__

# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use (Not defining
# uses the default style of the html_theme).
# pygments_style = 'sphinx'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# default language for highlighting in source code
highlight_language = "python3"

# show the code of plots that follows the command .. plot:: based on the
# package matplotlib.sphinxext.plot_directive
plot_include_source = True

# intersphinx mapping
intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'pyfar': ('https://pyfar.readthedocs.io/en/stable/', None),
    'gallery': ('https://pyfar-gallery.readthedocs.io/en/latest/', None),
    }

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']
html_css_files = ['css/custom.css']
html_js_files = ["js/custom.js"]
html_logo = 'resources/logos/pyfar_logos_fixed_size_spharpy.png'
html_title = "spharpy"
html_favicon = '_static/favicon.ico'

# -- HTML theme options
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html
html_sidebars = {
  "spharpy": []
}

html_theme_options = {
    "navbar_start": ["navbar-logo"],
    "navbar_end": ["navbar-icon-links", "theme-switcher"],
    "navbar_align": "content",
    "header_links_before_dropdown": None,  # will be automatically set later based on headers.rst
    "header_dropdown_text": "Packages",  # Change dropdown name from "More" to "Packages"
    "icon_links": [
        {
          "name": "GitHub",
          "url": "https://github.com/pyfar",
          "icon": "fa-brands fa-square-github",
          "type": "fontawesome",
        },
    ],
    # Configure secondary (right) side bar
    "show_toc_level": 3,  # Show all subsections of notebooks
    "secondary_sidebar_items": ["page-toc"],  # Omit 'show source' link that that shows notebook in json format
    "navigation_with_keys": True,
    # Configure navigation depth for section navigation
    "navigation_depth": 2,
}

html_context = {
   "default_mode": "light"
}

# redirect index to pyfar.html
redirects = {
     "index": "spharpy.html"
}

# -- download navbar and style files from gallery -----------------------------
branch = 'main'
link = f'https://github.com/pyfar/gallery/raw/{branch}/docs/'
folders_in = [
    '_static/css/custom.css',
    '_static/js/custom.js',
    '_static/favicon.ico',
    '_static/header.rst',
    'resources/logos/pyfar_logos_fixed_size_spharpy.png',
    ]

def download_files_from_gallery(link, folders_in):
    c = urllib3.PoolManager()
    for file in folders_in:
        url = link + file
        filename = file
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with c.request('GET', url, preload_content=False) as res:
            if res.status == 200:
                with open(filename, 'wb') as out_file:
                    shutil.copyfileobj(res, out_file)

download_files_from_gallery(link, folders_in)
# if logo does not exist, use pyfar logo
if not os.path.exists(html_logo):
    download_files_from_gallery(
        link, ['resources/logos/pyfar_logos_fixed_size_pyfar.png'])
    shutil.copyfile(
        'resources/logos/pyfar_logos_fixed_size_pyfar.png', html_logo)

# replace spharpy hard link to internal link
with open("_static/header.rst", "rt") as fin:
    with open("header.rst", "wt") as fout:
        lines = [line.replace(f'https://{project}.readthedocs.io', project) for line in fin]
        contains_project = any(project in line for line in lines)

        fout.writelines(lines)

        # add project to the list of projects if not in header
        if not contains_project:
            fout.write(f'   {project} <{project}>\n')

        # count the number of gallery headings
        count_gallery_headings = np.sum(
            ['https://pyfar-gallery.readthedocs.io' in line for line in lines])


# set dropdown header after gallery headings
html_theme_options['header_links_before_dropdown'] = count_gallery_headings+1





# Recursive Zonal Equal Area (EQ) Sphere Partitioning Toolbox.
# Release 1.10 2005-06-26
# COPYING
#
# For references, see AUTHORS.
# For revision history, see CHANGELOG.
#
# Copyright (c) 2004, 2005, University of New South Wales
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""
Python library partition the unit sphere into partition with equal area.
"""

from math import gcd
import numpy as np
import scipy.special as sps


def point_set(dimension, N):
    """Uniform distribution of points on the unit-sphere, by partitioning the
    unit-sphere into patches with equal area. This function is only a wrapper
    for point_set_polar and returns Cartesian coordinates instead of polar
    coordinates.

    Parameters
    ----------
    dimension : int
        The dimension, 2 represents is the 2-sphere and returns colatitude
        and azimuth angles.
    N : int
        The number of points to distribute.

    Returns
    -------
    points : array, double
        The resulting points in Cartesian coordinates. The dimensions of the
        array are (3, N)

    """
    dimension = np.asarray(dimension)
    N = np.asarray(N)
    return polar2cart(point_set_polar(dimension, N))


def point_set_polar(dimension, N):
    """Uniform distribution of points on the unit-sphere, by partitioning the
    unit-sphere into patches with equal area.

    Parameters
    ----------
    dimension : int
        The dimension, 2 represents is the 2-sphere and returns colatitude
        and azimuth angles.
    N : int
        The number of points to distribute.

    Returns
    -------
    points : array, double
        The resulting points in polar coordinates. The dimensions of the
        array are (dimension, N)

    """
    a_cap, n_regions = caps(dimension, N)

    if dimension == 1:
        points_s = a_cap - np.pi/N
    else:
        n_collars = np.size(n_regions) - 2

        points_s = np.zeros((dimension, N))
        point_n = 2

        offset = 0

        for collar_n in range(0, n_collars):
            # a_top is the colatitude of the top of the current collar.
            a_top = a_cap[collar_n]

            # a_bot is the colatitude of the bottom of the current collar.
            a_bot = a_cap[collar_n+1]

            # n_in_collar is the number of regions in the current collar.
            n_in_collar = n_regions[collar_n+1]

            points_l = point_set_polar(dimension - 1, n_in_collar)

            a_point = (a_top + a_bot)/2

            point_l_n = np.arange(0, np.size(points_l), dtype=int)

            if dimension == 2:
                points_s[0:dimension-1, point_n+point_l_n-1] = \
                    np.mod(points_l[point_l_n] + 2*np.pi*offset, 2*np.pi)

                offset += circle_offset(n_in_collar, n_regions[2+collar_n])
                offset -= np.floor(offset)
            else:
                points_s[0:dimension-2, point_n+point_l_n-1] = \
                    points_l[:, point_l_n]

            points_s[dimension-1, point_n+point_l_n-1] = a_point
            point_n += np.size(points_l)

        points_s[:, -1] = np.zeros(dimension)
        points_s[-1, -1] = np.pi

    return points_s


def caps(dimension, N):
    """Partition a sphere into two nested spherical caps.

    Does the following:
    1)  partitions the unit sphere S^dim into a list of spherical caps of
        increasing colatitude and thus increasing area,
    2)  sets S_CAP to be an array of size (1 by N_COLLARS+2),
        containing increasing colatitudes of caps, and
    3)  sets N_REGIONS to be an array of size (1 by N_COLLARS+2),
        containing the intger number of regions in each corresponding zone of
        S^dim.

    Examples
    --------
    % > [s_cap,n_regions] = eq_caps(2,10)
    % s_cap =
    %     0.6435    1.5708    2.4981    3.1416
    % n_regions =
    %      1     4     4     1
    %
    % > [s_cap,n_regions] = eq_caps(3,6)
    % s_cap =
    %     0.9845    2.1571    3.1416
    % n_regions =
    %      1     4     1
    %

    Parameters
    ----------
    dimension : int
        The dimension, 2 represents is the 2-sphere and returns colatitude
        and azimuth angles.
    N : int
        The number of points to distribute.

    Returns
    -------
    s_cap : array, double
        The colatitudes of the spherical caps. The dimensions of the array are
        (1, N_collars+2).
    n_regions : array, int
        The number of regions in each collar and the polar caps.
        The dimensions of the array are (1, N_collars+2).

    """
    if dimension == 1:
        s_cap = np.linspace(2*np.pi/N, 2*np.pi, N)
        n_regions = np.ones(10, dtype=int)
    elif N == 1:
        s_cap = np.pi
        n_regions = 1
    else:
        c_polar = polar_colat(dimension, N)
        n_collars = num_collars(N, c_polar, ideal_collar_angle(dimension, N))
        r_regions = ideal_region_list(dimension, N, c_polar, n_collars)
        n_regions = round_to_naturals(r_regions)
        s_cap = cap_colats(dimension, N, c_polar, n_regions)

    return s_cap, n_regions


def polar_colat(dimension, N):
    """
    The colatitude of the North polar (top) spherical cap
    Given dim and N, determine the colatitude of the North polar spherical cap.

    Parameters
    ----------
    dimension : int
        The dimension
    N : int
        The number of partitions

    Returns
    -------
    colatitude : double
        The colatitude angle of the top cap.
    """
    if N == 1:
        c_polar = np.pi
    elif N == 2:
        c_polar = np.pi/2
    else:
        ideal_region_area = area_of_ideal_region(dimension, N)
        c_polar = sradius_of_cap(dimension, ideal_region_area)

    return c_polar


def ideal_region_list(dimension, N, c_polar, n_collars):
    """The ideal real number of regions in each zone.

    List the ideal real number of regions in each collar, plus the polar caps.
    Given dim, N, c_polar and n_collars, determine r_regions, a list of the
    ideal real number of regions in each collar, plus the polar caps.
    The number of elements is n_collars+2.
    r_regions[1] is 1.
    r_regions[n_collars+2] is 1.
    The sum of r_regions is N.

    Parameters
    ----------
    dimension : int
        The dimension
    N : int
        The number of points
    c_polar : float
        The colatitude angle of the polar caps in radians. This defines the
        angular boundary between the north/south polar caps and the collar
        region. The north polar cap extends from 0 to c_polar, the collar
        region spans from c_polar to (π - c_polar), and the south polar cap
        extends from (π - c_polar) to π. Must be in the range [0, π/2].
    n_collars : int
        The number of collar elements

    Returns
    -------
    ideal_regions : double
        The ideal number of regions in a collar

    """
    r_regions = np.zeros(2+n_collars)
    r_regions[0] = 1

    if n_collars > 0:
        a_fitting = (np.pi - 2*c_polar) / n_collars
        ideal_region_area = area_of_ideal_region(dimension, N)
        for collar_n in range(1, n_collars+1):
            ideal_collar_area = area_of_collar(
                dimension,
                c_polar + (collar_n - 1) * a_fitting,
                c_polar + collar_n * a_fitting)
            r_regions[collar_n] = ideal_collar_area / ideal_region_area

    r_regions[-1] = 1

    return r_regions


def round_to_naturals(r_regions):
    """Round off a given list of numbers of regions
    Given N and r_regions, determine n_regions, a list of the natural number
    of regions in each collar and the polar caps.
    This list is as close as possible to r_regions, using rounding.
    The number of elements is n_collars+2.
    n_regions[1] is 1.
    n_regions[n_collars+2] is 1.
    The sum of n_regions is N.

    Parameters
    ----------
    r_regions : double
        The ideal number of regions per collar before rounding

    Returns
    -------
    n_regions : int
        The rounded integer number per collar.
    """
    r_regions = np.asarray(r_regions)
    n_regions = np.zeros(r_regions.shape, dtype=int)
    discrepancy = 0

    for zone_n in range(0, np.size(r_regions)):
        n_regions[zone_n] = np.rint(r_regions[zone_n] + discrepancy)
        discrepancy += (r_regions[zone_n] - n_regions[zone_n])

    return n_regions


def cap_colats(dimension, N, c_polar, n_regions):
    """Colatitudes of spherical caps enclosing cumulative sum of regions
    Given dim, N, c_polar and n_regions, determine c_caps, an increasing list
    of colatitudes of spherical caps which enclose the same area as that given
    by the cumulative sum of regions.
    The number of elements is n_collars+2.
    c_caps[1] is c_polar.
    c_caps[n_collars+1] is Pi-c_polar.
    c_caps[n_collars+2] is Pi.

    Parameters
    ----------
    dimension : int
        The dimension
    N : int
        Number of partitions
    c_polar : double
        Colatitude angles of the spherical caps
    n_regions: int
        Number of regions

    Returns
    -------
    c_caps : double
        Colatitude angles of the caps
    """
    c_caps = np.zeros(np.size(n_regions))
    c_caps[0] = c_polar
    ideal_region_area = area_of_ideal_region(dimension, N)
    n_collars = np.size(n_regions) - 2
    subtotal_n_regions = 1

    for collar_n in range(1, n_collars+1):
        subtotal_n_regions += n_regions[collar_n]
        c_caps[collar_n] = sradius_of_cap(
            dimension, subtotal_n_regions*ideal_region_area)

    c_caps[-1] = np.pi
    return c_caps


def num_collars(N, c_polar, a_ideal):
    """The number of collars between the polar caps
    Given N, an ideal angle, and c_polar, determine n_collars, the number of
    collars between the polar caps.

    Parameters
    ----------
    N : int
        The dimension
    c_polar : double
        The colatitude angle of the polar caps
    a_ideal : double
        The ideal collar angles.

    Returns
    -------
    n_collars : int
        The real number of collars.

    """
    if int((N > 2) and (a_ideal > 0)):
        n_collars = np.maximum(1, np.rint((np.pi - 2*c_polar) / a_ideal))
    else:
        n_collars = 0

    return int(n_collars)


def circle_offset(n_top, n_bot, extra_twist=False):
    """Try to maximize minimum distance of center points for S^2 collars
    Given n_top and n_bot, calculate an offset.
    The values n_top and n_bot represent the numbers of
    equally spaced points on two overlapping circles.
    The offset is given in multiples of whole rotations, and
    consists of three parts;
    1)  Half the difference between a twist of one sector on each of bottom
        and top. This brings the centre points into alignment.
    2)  A rotation which will maximize the minimum angle between
        points on the two circles.
    3)  An optional extra twist by a whole number of sectors on the second
        circle. The extra twist is added so that the location of the minimum
        angle between circles will progressively twist around the sphere
        with each collar.

    Parameters
    ----------
    n_top : int
        Number of points in the upper circle
    n_bot : int
        Number of points in the lower circle
    extra_twist : boolean
        Perform an additional rotation (see part 3)

    Returns
    -------
    offset : int
        The offset

    """
    offset = (1/n_bot - 1/n_top)/2 + gcd(n_top, n_bot) / (2*n_top*n_bot)
    if extra_twist:
        twist = 6
        offset += twist/n_bot

    return offset


def ideal_collar_angle(dimension, N):
    """Calculate the ideal angle for spherical collars.
    The ideal collar angle is determined by the side of a dim-dimensional
    hypercube of the same volume as the area of a single region of an N region
    equal area partition of S^dim.

    Since the EQ partition for N < 3 has no spherical collars,
    the recursive zonal equal area sphere partitioning algorithm does not use
    ideal_collar_angle(dim,N) for N < 3.

    dimension : int
        The dimension, 2 represents is the 2-sphere and returns colatitude
        and azimuth angles.
    N : int
        The number of points to distribute.

    Returns
    -------
    angle : double
        The ideal collar angle.

    """
    return area_of_ideal_region(dimension, N)**(1 / dimension)


def area_of_ideal_region(dimension, N):
    """The area a partition for an ideal partitioning of the sphere into
    N patches.

    Parameters
    ----------
    dimension : int
        The dimension
    N : int
        The number of partitions

    Returns
    -------
    area : double
        The area of an ideal partition.

    """
    return area_of_sphere(dimension)/N


def area_of_sphere(dimension):
    """Area of a sphere with the given dimension.

    Parameters
    ----------
    dimension : int
        Dimension of the sphere.

    Returns
    -------
    area : double
        The area.

    """
    power = (dimension + 1)/2
    return 2*np.pi**power / sps.gamma(power)


def area_of_collar(dimension, a_top, a_bot):
    """Area of spherical collar
    area_of_collar(dim, a_top, a_bot) calculates the area of
    an S^dim spherical collar specified by a_top, a_bot, where
    a_top is top (smaller) spherical radius,
    a_bot is bottom (larger) spherical radius.

    Parameters
    ----------
    dimension : int
        The dimension.
    a_top : array, double
        The top (smaller) spherical radius,
    a_bot : array, double
        The bottom (larger) spherical radius.

    Returns
    -------
    area : array, double
        The area of the spherical collar

    """
    return area_of_cap(dimension, a_bot) - area_of_cap(dimension, a_top)


def area_of_cap(dimension, s_cap):
    """Calculate the area of a spherical cap from the spherical radius s_cap.

    Parameters
    ----------
    dimension : int
        The dimension
    s_cap : double
        The spherical radius of the cap.

    Returns
    -------
    area : double
        The resulting area
    """
    if dimension == 1:
        area = 2 * s_cap
    elif dimension == 2:
        area = 4 * np.pi * np.sin(s_cap / 2)**2
    elif dimension == 3:
        raise ValueError("Dimension needs to be smaller or equal 2.")
    else:
        area = area_of_sphere(dimension) * sps.betainc(np.sin(s_cap/2)**2,
                                                       dimension/2,
                                                       dimension/2)

    return area


def sradius_of_cap(dimension, area):
    """Spherical radius of a spherical cap of a given area. It is assumed to be
    in the range [0, pi].
    The area is defined via the Lebesgue measure on S^dim inherited from
    its embedding in R^(dim+1).
    For dim <= 2, the spherical radius is calculated in closed form.

    Ref: [LeGS01 Lemma 4.1 p255].

    > s_cap=sradius_of_cap(2,area_of_sphere(2)/2)

    s_cap =
        1.5708

    s_cap=sradius_of_cap(3,(0:4)*area_of_sphere(3)/4)
    s_cap =
        [0, 1.1549, 1.5708, 1.9867, 3.1416]

    Parameters
    ----------
    dimension : int
        The dimension
    area : double
        The area of the spherical cap

    Returns
    -------
    s_rad : double
        The spherical radius

    Example
    -------
    >>> s_cap = sradius_of_cap(3,(0:4)*area_of_sphere(3)/4)
    >>> s_cap=sradius_of_cap(3,(0:4)*area_of_sphere(3)/4)

    """
    if dimension == 1:
        s_cap = area/2
    elif dimension == 2:
        s_cap = 2*np.arcsin(np.sqrt(area / np.pi) / 2)
    else:
        raise NotImplementedError
    return np.asarray(s_cap)


def polar2cart(points_polar):
    """Conversion from the polar angles theta and phi to Cartesian coordinates.

        x = cos(phi) * sin(theta)
        y = sin(phi) * sin(theta)
        x = cos(theta)

    Parameters
    ----------
    points_polar : array, double
        The points in polar coordinates, with shape (2, N)

    Returns
    -------
    points_cart : array, double
        The points in Cartesian coordinates with shape (3, N)

    """
    points_cart = np.zeros((points_polar.shape[0]+1, points_polar.shape[1]))
    points_cart[0, :] = np.cos(points_polar[0, :]) * np.sin(points_polar[1, :])
    points_cart[1, :] = np.sin(points_polar[0, :]) * np.sin(points_polar[1, :])
    points_cart[2, :] = np.cos(points_polar[1, :])
    return points_cart



r"""
The SamplingSphere class inherits from the :py:mod:`pyfar.classes.coordinates`
class, which supports various coordinate systems and the conversion between
them.
:py:class:`~spharpy.SamplingSphere` is designed to represent a set of
points on a sphere.

Therefore, all points must have the same radius within an absolute tolerance,
defined by :py:attr:`~spharpy.SamplingSphere.radius_tolerance`. If the
:py:attr:`~spharpy.SamplingSphere.weights` are not None, their sum must
equal the integral over the unit sphere, which is :math:`4\pi`.

The property :py:attr:`~spharpy.SamplingSphere.quadrature` is relevant for
spherical harmonic processing. It specifies if the points form a
quadrature rule, which requires valid weights, that the maximum
spherical harmonic order of the sampling
grid :py:attr:`~spharpy.SamplingSphere.n_max` is specified, and the inner
product of the weighted spherical harmonics matrix :math:`\mathrm{Y}`
yields the identity matrix
:math:`\mathrm{Y}^\mathrm{T} \text{diag}\{w\}\mathrm{Y}=\mathrm{I}`,
with the weights vector :math:`w`. The sampling is considered a valid
quadrature if the maximum absolute deviation of
:math:`\mathrm{Y}^\mathrm{T} \text{diag}\{w\}\mathrm{Y}` from :math:`I` is
smaller than the specified
:py:attr:`~spharpy.SamplingSphere.quadrature_tolerance`.

It also adds the additional property:

- :py:attr:`~spharpy.SamplingSphere.n_max`: the maximum spherical harmonic
  order of the sampling grid.

Note that the :py:mod:`spharpy.samplings` module provides a set of
predefined spherical sampling grids, which can be used to create a
:py:class:`spharpy.SamplingSphere` object.
"""

import numpy as np
from pyfar.classes.coordinates import sph2cart, cyl2cart
import pyfar as pf
from spharpy.spherical import spherical_harmonic_basis_real


class SamplingSphere(pf.Coordinates):
    """Class for samplings on a sphere."""

    def __init__(
            self, x=None, y=None, z=None, n_max=None, weights: np.array = None,
            comment: str = "", radius_tolerance=1e-6,
            quadrature_tolerance=1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for more information.

        Parameters
        ----------
        x : ndarray, number
            X coordinate of a right handed Cartesian coordinate system in
            meters (-\infty < x < \infty).
        y : ndarray, number
            Y coordinate of a right handed Cartesian coordinate system in
            meters (-\infty < y < \infty).
        z : ndarray, number
            Z coordinate of a right handed Cartesian coordinate system in
            meters (-\infty < z < \infty).
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid. If
            provided, it must be an integer greater or equal to 0.
            The default is ``None``.
        weights: array like, number, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.
        """
        self._radius_tolerance = None
        self.radius_tolerance = radius_tolerance

        pf.Coordinates.__init__(
            self, x, y, z, weights=weights, comment=comment)
        self._n_max = n_max

        self._quadrature_tolerance = None
        self.quadrature_tolerance = quadrature_tolerance

        self._quadrature = None

    @classmethod
    def from_coordinates(
            cls, coordinates, n_max=None, radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Convert Coordinates class object to SamplingSphere class object.

        Parameters
        ----------
        coordinates : pyfar.Coordinates
            Coordinates to be converted.
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.
        """

        if type(coordinates) is not pf.Coordinates:
            raise TypeError('coordinates must be a pyfar Coordinates object')

        # make sure mutable data is copied
        if coordinates.weights is None:
            weights = None
        else:
            weights = coordinates.weights.copy()

        return cls(
            coordinates.x, coordinates.y, coordinates.z,
            weights=weights, comment=coordinates.comment,
            n_max=n_max, radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @classmethod
    def from_cartesian(
            cls, x, y, z, n_max=None, weights: np.array = None,
            comment: str = "", radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for  more information.

        Parameters
        ----------
        x : ndarray, number
            X coordinate of a right handed Cartesian coordinate system in
            meters (-\infty < x < \infty).
        y : ndarray, number
            Y coordinate of a right handed Cartesian coordinate system in
            meters (-\infty < y < \infty).
        z : ndarray, number
            Z coordinate of a right handed Cartesian coordinate system in
            meters (-\infty < z < \infty).
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        weights: array like, number, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.

        Examples
        --------
        Create a SamplingSphere object

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere.from_cartesian(0, 0, 1)

        or the same using

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere(0, 0, 1)
        """
        return cls(
            x, y, z, weights=weights, comment=comment, n_max=n_max,
            radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @classmethod
    def from_spherical_elevation(
            cls, azimuth, elevation, radius, n_max=None,
            weights: np.array = None, comment: str = "",
            radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for  more information.

        Parameters
        ----------
        azimuth : ndarray, double
            Angle in radiant of rotation from the x-y-plane facing towards
            positive x direction. Used for spherical and cylindrical coordinate
            systems.
        elevation : ndarray, double
            Angle in radiant with respect to horizontal plane (x-z-plane).
            Used for spherical coordinate systems.
        radius : ndarray, double
            Distance to origin for each point. Used for spherical coordinate
            systems.
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        weights: array like, float, None, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.

        Examples
        --------
        Create a SamplingSphere object

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere.from_spherical_elevation(0, 0, 1)
        """

        x, y, z = sph2cart(
            azimuth, np.pi / 2 - np.atleast_1d(elevation), radius)
        return cls(
            x, y, z, weights=weights, comment=comment, n_max=n_max,
            radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @classmethod
    def from_spherical_colatitude(
            cls, azimuth, colatitude, radius, n_max=None,
            weights: np.array = None, comment: str = "",
            radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for  more information.

        Parameters
        ----------
        azimuth : ndarray, double
            Angle in radiant of rotation from the x-y-plane facing towards
            positive x direction. Used for spherical and cylindrical coordinate
            systems.
        colatitude : ndarray, double
            Angle in radiant with respect to polar axis (z-axis). Used for
            spherical coordinate systems.
        radius : ndarray, double
            Distance to origin for each point. Used for spherical coordinate
            systems.
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        weights: array like, number, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.

        Examples
        --------
        Create a SamplingSphere object

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere.from_spherical_colatitude(
        ...     0, 0, 1)
        """

        x, y, z = sph2cart(azimuth, colatitude, radius)
        return cls(
            x, y, z, weights=weights, comment=comment, n_max=n_max,
            radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @classmethod
    def from_spherical_side(
            cls, lateral, polar, radius, n_max=None,
            weights: np.array = None, comment: str = "",
            radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for  more information.

        Parameters
        ----------
        lateral : ndarray, double
            Angle in radiant with respect to horizontal plane (x-y-plane).
            Used for spherical coordinate systems.
        polar : ndarray, double
            Angle in radiant of rotation from the x-z-plane facing towards
            positive x direction. Used for spherical coordinate systems.
        radius : ndarray, double
            Distance to origin for each point. Used for spherical coordinate
            systems.
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        weights: array like, number, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.

        Examples
        --------
        Create a SamplingSphere object

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere.from_spherical_side(0, 0, 1)
        """

        x, z, y = sph2cart(
            polar, np.pi / 2 - np.atleast_1d(lateral), radius)
        return cls(
            x, y, z, weights=weights, comment=comment, n_max=n_max,
            radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @classmethod
    def from_spherical_front(
            cls, frontal, upper, radius, n_max=None, weights: np.array = None,
            comment: str = "", radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for  more information.

        Parameters
        ----------
        frontal : ndarray, double
            Angle in radiant of rotation from the y-z-plane facing towards
            positive y direction. Used for spherical coordinate systems.
        upper : ndarray, double
            Angle in radiant with respect to polar axis (x-axis). Used for
            spherical coordinate systems.
        radius : ndarray, double
            Distance to origin for each point. Used for spherical coordinate
            systems.
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        weights: array like, number, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.

        Examples
        --------
        Create a SamplingSphere object

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere.from_spherical_front(0, 0, 1)
        """

        y, z, x = sph2cart(frontal, upper, radius)
        return cls(
            x, y, z, weights=weights, comment=comment, n_max=n_max,
            radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @classmethod
    def from_cylindrical(
            cls, azimuth, z, rho, n_max=None, weights: np.array = None,
            comment: str = "", radius_tolerance: float = 1e-6,
            quadrature_tolerance: float = 1e-10):
        r"""
        Create a SamplingSphere class object from a set of points on a sphere.

        See :py:mod:`pyfar.classes.coordinates` for  more information.

        Parameters
        ----------
        azimuth : ndarray, double
            Angle in radiant of rotation from the x-y-plane facing towards
            positive x direction. Used for spherical and cylindrical coordinate
            systems.
        z : ndarray, double
            The z coordinate
        rho : ndarray, double
            Distance to origin for each point in the x-y-plane. Used for
            cylindrical coordinate systems.
        n_max : int, optional
            Maximum spherical harmonic order of the sampling grid.
            The default is ``None``.
        weights: array like, number, optional
            Weighting factors for coordinate points. Their sum must equal to
            the integral over the unit sphere, which is :math:`4\pi`.
            The `shape` of the array must match the `shape` of the individual
            coordinate arrays. The default is ``None``, which means that no
            weights are used.
        comment : str, optional
            Comment about the stored coordinate points. The default is
            ``""``, which initializes an empty string.
        radius_tolerance : float, optional
            All points that are stored in a SamplingSphere must have the same
            radius and an error is raised if the maximum deviation from the
            mean radius exceeds this tolerance. The default of ``1e-6`` meter
            is intended to allow for some numerical inaccuracy.
        quadrature_tolerance : float, optional
            Tolerance for testing whether the provided sampling grid is
            a valid quadrature. The sampling is considered a valid quadrature
            if the maximum absolute deviation of the inner product of the
            weighted spherical harmonics matrix from the identity matrix is
            smaller than the specified tolerance. The default is ``1e-10``.

        Examples
        --------
        Create a SamplingSphere object

        >>> import spharpy
        >>> sampling = spharpy.SamplingSphere.from_cylindrical(
        ...     0, 0, 1, sh_order=1)
        """

        x, y, z = cyl2cart(azimuth, z, rho)
        return cls(
            x, y, z, weights=weights, comment=comment, n_max=n_max,
            radius_tolerance=radius_tolerance,
            quadrature_tolerance=quadrature_tolerance)

    @property
    def n_max(self):
        """Get or set the spherical harmonic order."""
        return self._n_max

    @n_max.setter
    def n_max(self, value):
        """Get or set the spherical harmonic order."""
        assert value >= 0

        if value is None:
            self._n_max = None
            self._quadrature = False
        else:
            if self._n_max != value:
                self._quadrature = None
            self._n_max = int(value)

    @property
    def radius_tolerance(self):
        """Get or set the radius tolerance in meter."""
        return self._radius_tolerance

    @radius_tolerance.setter
    def radius_tolerance(self, value):
        """Get or set the radius tolerance in meter."""

        # check input
        if not isinstance(value, (int, float)) or value <= 0:
            raise ValueError(
                'The radius tolerance must be a number greater than zero')

        current_tolerance = self.radius_tolerance
        self._radius_tolerance = float(value)

        # Check if points meet new tolerance if points exist
        if hasattr(self, 'x'):
            try:
                self._check_points(self._x, self._y, self._z)
            except ValueError as e:
                # revert setting the tolerance and raise the error
                self._radius_tolerance = current_tolerance
                raise e

    @property
    def quadrature_tolerance(self):
        """Get or set the quadrature tolerance."""
        return self._quadrature_tolerance

    @quadrature_tolerance.setter
    def quadrature_tolerance(self, value):
        """Get or set the quadrature tolerance."""

        # check input
        if not isinstance(value, (int, float)) or value <= 0:
            raise ValueError(
                'The quadrature tolerance must be a number greater than zero')

        self._quadrature_tolerance = float(value)
        self._quadrature = None

    def _check_points(self, x, y, z):
        """Check input data before setting coordinates."""

        # convert to numpy arrays of the same shape
        x, y, z = super()._check_points(x, y, z)

        # check for equal radius
        radius = np.sqrt(x.flatten()**2 + y.flatten()**2 + z.flatten()**2)
        radius_delta = np.max(radius) - np.min(radius)
        if radius_delta > self.radius_tolerance:
            raise ValueError(
                'All points must have the same radius but the difference '
                f'between the minimum and maximum radius is {radius_delta:.3g}'
                ' m, which exceeds the tolerance of '
                f'{self.radius_tolerance:.3g} m. The tolerance can be changed '
                'using SamplingSphere.radius_tolerance.')

        # reset the quadrature flag to make sure it is checked again after
        # adding or changing points in the SamplingSphere
        self._quadrature = None

        return x, y, z

    def _check_weights(self, weights):
        r"""Check if the weights are valid.
        The weights must be positive and their sum must equal integration of
        the unit sphere, i.e. :math:`4\pi`.

        Parameters
        ----------
        weights : array like, number
            the weights for each point, should be of size of self.csize.

        Returns
        -------
        weights : np.ndarray[float64], None
            The weights reshaped to the cshape of the coordinates if not None.
            Otherwise None.
        """
        weights = super()._check_weights(weights)

        if weights is None:
            return weights
        if np.any(weights < 0) or np.any(np.isnan(weights)):
            raise ValueError("All weights must be positive numeric values.")

        if not np.isclose(np.sum(weights), 4*np.pi, atol=1e-6, rtol=1e-6):
            raise ValueError(
                "The sum of the weights must be equal to 4*pi. "
                f"Current sum: {np.sum(weights)}")

        return weights

    def _check_quadrature(self):
        r"""Check if the sampling is a valid quadrature.

        Returns
        -------
        check : bool
            Indicates if sampling is a valid quadrature
        """
        if self.n_max is None or self.weights is None:
            return False
        # create basis matrix
        sh_basis = spherical_harmonic_basis_real(self.n_max, self)

        # test if basis is quadrature
        quad_evaluation = sh_basis.T @ np.diag(self.weights) @ sh_basis
        identity = np.eye((self.n_max + 1)**2)

        error = np.max(np.abs(quad_evaluation-identity))
        return error < self.quadrature_tolerance

    @property
    def weights(self):
        r"""The area/quadrature weights of the sampling.
        Their sum must equal to :math:`4\pi`.
        """
        return super().weights

    @weights.setter
    def weights(self, weights):
        r"""Get or set the area/quadrature weights of the sampling.
        Their sum must equal to :math:`4\pi`.
        """
        if not np.array_equal(weights, self.weights):
            self._quadrature = None

        super(__class__, type(self)).weights.fset(self, weights)

    @property
    def quadrature(self):
        """Get the quadrature flag."""
        if self._quadrature is None:
            # recompute _quadrature flag
            self._quadrature = self._check_quadrature()

        return self._quadrature


    def __repr__(self):
        """Get info about SamplingSphere object."""
        if self.csize == 0:
            return 'SamplingSphere: empty'

        return f'SamplingSphere: n_max={self.n_max}, cshape={self.cshape}'



import os
import pytest
import spharpy as sp
from pyfar.testing.plot_utils import create_figure, save_and_compare
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import re

"""
For general information on testing plot functions see
https://pyfar-gallery.readthedocs.io/en/latest/contribute/contribution_guidelines.html#testing-plot-functions

Important:
- `create_baseline` and `compare_output` must be ``False`` when pushing
  changes to pyfar.
- `create_baseline` must only be ``True`` if the behavior of a plot function
  changed. In this case it is best practice to recreate only the baseline plots
  of the plot function (plot behavior) that changed.
"""
# global parameters -----------------------------------------------------------
create_baseline = False

# file type used for saving the plots
file_type = "png"

# if true, the plots will be compared to the baseline and an error is raised
# if there are any differences. In any case, differences are written to
# output_path as images
compare_output = False

# path handling
base_path = os.path.join('tests', 'test_plot_data')
baseline_path = os.path.join(base_path, 'baseline')
output_path = os.path.join(base_path, 'output')

if not os.path.isdir(base_path):
    os.mkdir(base_path)
if not os.path.isdir(baseline_path):
    os.mkdir(baseline_path)
if not os.path.isdir(output_path):
    os.mkdir(output_path)

# remove old output files
for file in os.listdir(output_path):
    os.remove(os.path.join(output_path, file))

# the naming scheme of the baseline is as follows:
# <function_name>_<parameter_name>_<parameters>.png

# testing ---------------------------------------------------------------------
@pytest.fixture(autouse=True)
def _close_all_figures():
    """
    Close all matplotlib figures after each test to prevent test
    pollution.
    """
    yield
    plt.close('all')


@pytest.mark.parametrize('function', [
    (sp.plot.scatter),
    (sp.plot.voronoi_cells_sphere)])
def test_sampling_scatter(function):
    """Test the scatter plot with default arguments."""
    coords = sp.samplings.equal_area(n_max=0, n_points=12)

    # do plotting
    filename = f'{function.__name__}_default'
    create_figure()
    function(coords)
    save_and_compare(
        create_baseline, baseline_path, output_path, filename,
        file_type, compare_output)

    match = 'must be a coordinates object.'
    with pytest.raises(ValueError, match=match):
        function('coords')


@pytest.mark.parametrize('function', [
    (sp.plot.balloon),
    (sp.plot.balloon_wireframe),
    (sp.plot.contour),
    (sp.plot.contour_map),
    (sp.plot.pcolor_map),
    (sp.plot.pcolor_sphere)])
def test_spherical_default(function):
    """Test all spherical plots with default arguments."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # do plotting
    filename = f'{function.__name__}_default'
    create_figure()
    function(coords, data)
    save_and_compare(
        create_baseline, baseline_path, output_path, filename,
        file_type, compare_output)

    # test error for invalid inputs
    match = 'data must be a 1D array with the same cshape as the coordinates'
    with pytest.raises(ValueError, match=match):
        function(coords, 'data')

    match = 'coordinates must be a coordinates object.'
    with pytest.raises(ValueError, match=match):
        function('coords', data)


@pytest.mark.parametrize('function', [
    (sp.plot.contour),
    (sp.plot.contour_map),
    (sp.plot.pcolor_map)])
@pytest.mark.parametrize('cmap', [plt.get_cmap('plasma'), 'plasma', None])
def test_spherical_cmap(function, cmap):
    """Test all spherical plots with custom cmap argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # do plotting
    if isinstance(cmap, mpl.colors.Colormap):
        # if a colormap object is passed, use its name for the filename
        cmap_str = 'ColormapObject'
    else:
        # otherwise use the string representation of the colormap
        cmap_str = str(cmap)
    filename = f'{function.__name__}_cmap_{cmap_str}'
    create_figure()
    function(coords, data, cmap=cmap)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)

    # test error for invalid inputs
    match = 'cmap'
    with pytest.raises(ValueError, match=match):
        function(coords, data, cmap=5)


@pytest.mark.parametrize('function', [
    (sp.plot.balloon),
    (sp.plot.balloon_wireframe),
    (sp.plot.contour),
    (sp.plot.contour_map),
    (sp.plot.pcolor_map),
    (sp.plot.pcolor_sphere)])
@pytest.mark.parametrize('colorbar', [True, False])
def test_spherical_colorbar(function, colorbar):
    """Test all spherical plots with custom colorbar argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    out = function(coords, data, colorbar=colorbar)
    if colorbar:
        assert isinstance(out[2], mpl.colorbar.Colorbar)
        assert isinstance(out[0], list)
    if not colorbar:
        assert out[2] is None
        assert isinstance(out[0], plt.Axes)

    # do plotting
    filename = f'{function.__name__}_colorbar_{colorbar}'
    create_figure()
    function(coords, data, colorbar=colorbar)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)

    # test error for invalid inputs
    match = 'colorbar must be a boolean.'
    with pytest.raises(ValueError, match=match):
        function(coords, data, colorbar='colorbar')


@pytest.mark.parametrize('function', [
    (sp.plot.balloon),
    (sp.plot.balloon_wireframe),
    (sp.plot.contour),
    (sp.plot.contour_map),
    (sp.plot.pcolor_map),
    (sp.plot.pcolor_sphere)])
@pytest.mark.parametrize('limits', [None, (-1.5, 1.5), [-1.5, 1.5]])
def test_spherical_limits(function, limits):
    """Test all spherical plots with custom limits argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # do plotting
    if limits is None:
        limits_str = 'None'
    else:
        limits_str = f'{type(limits).__name__}'
    filename = f'{function.__name__}_limits_{limits_str}'
    create_figure()
    function(coords, data, limits=limits)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)

    # test error for invalid inputs
    match = (
        "limits must be a tuple or list containing the minimum and "
        "maximum values for the colormap or None.")
    with pytest.raises(ValueError, match=match):
        function(coords, data, limits='limits')
    with pytest.raises(ValueError, match=match):
        function(coords, data, limits=[0, 1, 2])


@pytest.mark.parametrize('function', [
    (sp.plot.balloon),
    (sp.plot.balloon_wireframe),
    (sp.plot.pcolor_sphere)])
@pytest.mark.parametrize('cmap_encoding', ['phase', 'magnitude'])
@pytest.mark.parametrize('cmap', [None, 'magma'])
def test_balloon_cmap_encoding(function, cmap_encoding, cmap):
    """Test all balloon like plots with custom arguments."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # use the azimuth angle for a continuous phase change
    data = np.abs(data) * np.exp(1j * coords.azimuth)

    # do plotting
    filename = f'{function.__name__}_cmap_encoding_{cmap_encoding}_cmap_{cmap}'
    create_figure()
    function(coords, data, cmap_encoding=cmap_encoding, cmap=cmap)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)

@pytest.mark.parametrize('function', [
    (sp.plot.balloon),
    (sp.plot.balloon_wireframe),
    (sp.plot.pcolor_sphere)])
def test_balloon_cmap_encoding_error(function):
    """Test raising errors for invalid cmap_encoding argument."""
    # test error for invalid inputs
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    match = "cmap_encoding must be either 'phase' or 'magnitude'."
    with pytest.raises(ValueError, match=match):
        function(coords, data, cmap_encoding='invalid')


@pytest.mark.parametrize('function', [
    (sp.plot.pcolor_map),
    (sp.plot.contour_map),
    ])
@pytest.mark.parametrize('projection', ['mollweide', 'hammer'])
def test_spherical_projection(function, projection):
    """Test spherical plots with custom projection argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # do plotting
    filename = f'{function.__name__}_projection_{projection}'
    create_figure()
    function(coords, data, projection=projection)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)

    # test error for invalid inputs
    match = "The projection of the axis needs to be 'projection'"
    with pytest.raises(ValueError, match=match):
        function(coords, data, projection='projection')


@pytest.mark.parametrize('function', [
    sp.plot.contour_map,
    sp.plot.contour,
    ])
@pytest.mark.parametrize('levels', [
    8,
    (-1, -0.5, 0, 0.5, 1),
    [-1, -0.5, 0, 0.5, 1],
    np.array((-1, -0.5, 0, 0.5, 1)),
    ])
def test_spherical_levels(function, levels):
    """Test contour plots with custom level argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # do plotting
    filename = f'{function.__name__}_levels_{type(levels).__name__}'
    create_figure()
    function(coords, data, levels=levels)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)

    # test error for invalid inputs
    match = 'levels'
    with pytest.raises(ValueError, match=match):
        function(coords, data, levels='levels')


@pytest.mark.parametrize('function', [
    sp.plot.pcolor_map,
    ])
@pytest.mark.parametrize('refine', [
    True, False,
    ])
def test_spherical_refine(function, refine):
    """Test contour plots with custom refine argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    # do plotting
    filename = f'{function.__name__}_refine_{refine}'
    create_figure()
    function(coords, data, refine=refine)
    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)


@pytest.mark.parametrize('function', [
    sp.plot.pcolor_map,
    ])
def test_spherical_refine_error(function):
    """Test contour plots with invalid refine argument."""
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)

    match = 'refine'
    with pytest.raises(ValueError, match=match):
        function(coords, data, refine='refine')


@pytest.mark.parametrize(("function", "projection"), [
    (sp.plot.balloon, '3d'),
    (sp.plot.balloon_wireframe, '3d'),
    (sp.plot.pcolor_sphere, '3d'),
    (sp.plot.contour, 'rectilinear'),
    (sp.plot.contour_map, 'mollweide'),
    (sp.plot.pcolor_map, 'mollweide'),
    ])
def test_data_plots_projection_input_and_return(function, projection):
    """
    Test all spherical plots with ax argument and check if this is
    also returned.
    """
    coords = sp.samplings.equal_area(n_max=0, n_points=500)
    data = np.sin(coords.colatitude) * np.cos(coords.azimuth)
    # do plotting
    create_figure()
    ax = plt.axes(projection=projection)

    ax_out = function(coords, data, ax=ax)[0]

    # check if the returned axis has the correct projection
    assert ax_out[0].name == projection

    # test error for invalid inputs
    match = f"The projection of the axis needs to be '{projection}'"
    with pytest.raises(ValueError, match=match):
        function(coords, data, ax=plt.axes(projection='polar'))

    match = re.escape("If [ax1, ax2] is passed ax2 needs to be of"
                      " 'rectilinear' projection")
    with pytest.raises(ValueError, match=match):
        function(coords, data, ax=[
            plt.axes(projection=projection),
            plt.axes(projection='3d')])

@pytest.mark.parametrize('function', [
    (sp.plot.scatter),
    (sp.plot.voronoi_cells_sphere),
    ])
def test_coordinates_plots_projection_input_and_return(function):
    """Test scatter plots with ax argument and check if this is also returned.
    """
    coords = sp.samplings.equal_area(n_max=0, n_points=12)
    # do plotting
    create_figure()
    ax = plt.axes(projection='3d')

    ax_out = function(coords, ax=ax)

    # check if the returned axis is a 3D axis
    assert ax_out.name == '3d'

    # test error for invalid inputs
    match = "The projection of the axis needs to be '3d'"
    with pytest.raises(ValueError, match=match):
        function(coords, ax=plt.axes())


def test_cmap_phase_twilight():
    """Test the phase twilight colormap."""
    lut = 512
    filename = 'cmap_phase_twilight'
    create_figure(6, .5)

    # do plotting
    colormap = sp.plot.phase_twilight(lut)
    ax = plt.gca()
    gradient = np.linspace(0, 1, lut)
    gradient = np.vstack((gradient, gradient))
    ax.imshow(gradient, aspect='auto', cmap=colormap)
    ax.set_axis_off()

    save_and_compare(
        create_baseline, baseline_path, output_path, filename,
        file_type, compare_output)

    match = 'lut must be a positive integer.'
    with pytest.raises(ValueError, match=match):
        sp.plot.phase_twilight('lut')

    with pytest.raises(ValueError, match=match):
        sp.plot.phase_twilight(0)


@pytest.mark.parametrize(
        ("function"),
        [(sp.plot.balloon),
         (sp.plot.balloon_wireframe),
         (sp.plot.contour),
         (sp.plot.contour_map),
         (sp.plot.pcolor_sphere),
         (sp.plot.pcolor_map)],
)
@pytest.mark.parametrize(
    ("ax"), [(1), ([1, 2, 3]), np.array([1, 2, 3])],
)
def test_ax_parameter_errors(function, ax, equal_area_sampling):
    """Test all warnings related to `ax` parameter."""
    coords, data = equal_area_sampling
    # Test case when `ax`` is wrong type
    match = "ax can be ``None``, a single matplotlib.axes.Axes object or a " \
            "list, tuple or array of two axes"
    with pytest.raises(ValueError, match=match):
        function(coords, data, ax=ax)


@pytest.mark.parametrize(
        ("function"),
        [(sp.plot.balloon),
         (sp.plot.balloon_wireframe),
         (sp.plot.contour),
         (sp.plot.contour_map),
         (sp.plot.pcolor_sphere),
         (sp.plot.pcolor_map)],
)
def test_colorbar_ax_error(function, equal_area_sampling):
    """Test error raised by false colorbar / ax parameter combination."""
    coords, data = equal_area_sampling
    match = "A list of axes can not be used if colorbar is False"
    with pytest.raises(
        ValueError, match=match,
    ):
        function(coords, data, colorbar=False, ax=[plt.gca(), plt.gca()])


@pytest.mark.parametrize(
        ('function', 'projection'),
        [(sp.plot.balloon, '3d'),
         (sp.plot.balloon_wireframe, '3d'),
         (sp.plot.contour, 'rectilinear'),
         (sp.plot.contour_map, 'mollweide'),
         (sp.plot.pcolor_sphere, '3d'),
         (sp.plot.pcolor_map, 'mollweide')],
)
@pytest.mark.parametrize(
    ('ax_option'),
    [('none'), ('single'), ('two')],
)
def test_ax_input(function, projection, ax_option, equal_area_sampling):
    """Test ax input for None, single axis and list of two axes."""
    coords, data = equal_area_sampling

    filename = f'{function.__name__}_ax_{ax_option}'

    if ax_option == 'none':
        function(coords, data)
    if ax_option == 'single':
        ax1 = plt.subplot(111, projection=projection)
        function(coords, data, ax=ax1)
    if ax_option == 'two':
        ax1 = plt.subplot(121, projection=projection)
        ax2 = plt.subplot(122, projection='rectilinear')

        function(coords, data, ax=[ax1, ax2])

        plt.tight_layout()
        plt.subplots_adjust(wspace=0.2)

    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)


@pytest.mark.parametrize(
        ('function', 'kind'),
        [(sp.plot.scatter, 'coordinate_plot'),
         (sp.plot.voronoi_cells_sphere, 'coordinate_plot'),
         (sp.plot.pcolor_sphere, 'data_plot'),
         (sp.plot.pcolor_map, 'data_plot'),
         (sp.plot.balloon, 'data_plot'),
         (sp.plot.balloon_wireframe, 'data_plot'),
         (sp.plot.contour, 'data_plot'),
         (sp.plot.contour_map, 'data_plot')],
)
@pytest.mark.parametrize(
        'style',
        [('light'), ('dark')],
)
def test_pyfar_styles(function, style, kind, equal_area_sampling):
    """Test style-parameter with pyfar plot styles 'light' and 'dark'."""
    coords, data = equal_area_sampling
    filename = f'{function.__name__}_style_{style}'

    if kind == 'coordinate_plot':
        function(coords, style=style)
    elif kind == 'data_plot':
        function(coords, data, style=style)

    save_and_compare(create_baseline, baseline_path, output_path, filename,
                     file_type, compare_output)


@pytest.mark.parametrize(
        ('function', 'kind'),
        [(sp.plot.scatter, 'coordinate_plot'),
         (sp.plot.voronoi_cells_sphere, 'coordinate_plot'),
         (sp.plot.pcolor_sphere, 'data_plot'),
         (sp.plot.pcolor_map, 'data_plot'),
         (sp.plot.balloon, 'data_plot'),
         (sp.plot.balloon_wireframe, 'data_plot'),
         (sp.plot.contour, 'data_plot'),
         (sp.plot.contour_map, 'data_plot')],
)
def test_custom_style(function, kind, equal_area_sampling):
    """Test style parameter with custom value."""
    coords, data = equal_area_sampling
    style = {'axes.facecolor': '#000000'}

    if kind == 'coordinate_plot':
        function(coords, style=style)
    elif kind == 'data_plot':
        function(coords, data, style=style)

    facecolor = mpl.colors.to_hex(plt.gca().patch.get_facecolor())
    assert facecolor == style['axes.facecolor']



import pytest
import pyfar as pf
import numpy as np
from spharpy import SamplingSphere
from spharpy.samplings import equal_area


@pytest.fixture
def make_coordinates():
    """Fixture factory which can be used to return coordinate storing objects.

    Works with every child class of pyfar.Coordinates which supports the
    from_spherical_colatitude method. The default implementation is
    pyfar.Coordinates.
    """
    class Factory:
        @staticmethod
        def create_coordinates(
                implementation=pf.Coordinates,
                phi=np.pi/2, theta=np.pi/2, rad=1,
            ):
            return implementation.from_spherical_colatitude(phi, theta, rad)
    return Factory


def icosahedron_points():
    """Return the coordinate points of an icosahedron in spherical coordinates.

    Returns
    -------
    rad : float, ndarray
        The radius in meters.
    theta : float, ndarray
        The colatitude angle in radians.
    phi : float, ndarray
        The azimuth angle in radians.
    """
    gamma_R_r = np.arccos(np.cos(np.pi/3) / np.sin(np.pi/5))
    gamma_R_rho = np.arccos(1/(np.tan(np.pi/5) * np.tan(np.pi/3)))

    theta = np.tile(np.array([
        np.pi - gamma_R_rho,
        np.pi - gamma_R_rho - 2*gamma_R_r,
        2*gamma_R_r + gamma_R_rho,
        gamma_R_rho]), 5)
    theta = np.sort(theta)
    phi = np.arange(0, 2*np.pi, 2*np.pi/5)
    phi = np.concatenate((np.tile(phi, 2), np.tile(phi + np.pi/5, 2)))
    rad = np.ones(20)

    return rad, theta, phi


@pytest.fixture
def icosahedron():
    """Return the coordinate points of an icosahedron in spherical coordinates.

    Returns
    -------
    rad : float, ndarray
        The radius in meters.
    theta : float, ndarray
        The colatitude angle in radians.
    phi : float, ndarray
        The azimuth angle in radians.
    """
    rad, theta, phi = icosahedron_points()
    return rad, theta, phi


@pytest.fixture
def icosahedron_sampling():
    """Return the coordinate points of an icosahedron in spherical coordinates.

    Returns
    -------
    rad : float, ndarray
        The radius in meters.
    theta : float, ndarray
        The colatitude angle in radians.
    phi : float, ndarray
        The azimuth angle in radians.
    """
    rad, theta, phi = icosahedron_points()
    weights = np.ones_like(rad) * 4 * np.pi / len(rad)
    return SamplingSphere.from_spherical_colatitude(
        phi, theta, rad, weights=weights, n_max=2)


@pytest.fixture
def download_sampling():
    def download_sampling(kind, degree):
        if kind in ['extremal', 'hyperinterpolation']:
            from spharpy.samplings.samplings import _sph_extremal_load_data
            return _sph_extremal_load_data(degree)
        elif kind == 't-design':
            from spharpy.samplings.samplings import _sph_t_design_load_data
            return _sph_t_design_load_data(degree)

    return download_sampling


@pytest.fixture
def equal_area_sampling():
    """Return 500 point equal area test data."""
    coords = equal_area(n_max=0, n_points=500)
    data = np.sin(coords.azimuth) * np.cos(coords.elevation)
    return coords, data



import spharpy
from pyfar import Coordinates
import spharpy.interpolate as interpolate
import numpy as np


def test_smooth_sphere_bivariate_spline_interpolation(download_sampling):
    n_max = 10
    sampling = spharpy.samplings.equal_area(
        n_max, n_points=500, condition_num=np.inf)
    Y = spharpy.spherical.spherical_harmonic_basis_real(n_max, sampling)
    y_vec = spharpy.spherical.spherical_harmonic_basis_real(
        n_max, Coordinates(1, 0, 0))
    data = Y @ y_vec.T

    data = np.sin(sampling.colatitude)*np.sin(2*sampling.azimuth)

    download_sampling('extremal', 35)
    interp_grid = spharpy.samplings.hyperinterpolation(n_max=35)

    data_grid = np.sin(interp_grid.colatitude)*np.sin(2*interp_grid.azimuth)

    interpolator = interpolate.SmoothSphereBivariateSpline(
        sampling, data, s=1e-4)

    interp_data = interpolator(interp_grid)

    # check if error over entire sphere sufficiently small
    assert np.linalg.norm(np.abs(interp_data - data_grid)) / \
        np.linalg.norm(np.abs(data_grid)) < 1e-2



"""
Tests for spherical harmonic class.
"""
import pytest
import numpy as np
from spharpy import (
    SphericalHarmonics, SamplingSphere, SphericalHarmonicDefinition,
)
import pyfar as pf
import deepdiff


def test_spherical_harmonics_definition_init():
    """Test default behavior."""
    definition = SphericalHarmonicDefinition()
    assert definition.basis_type == 'real'
    assert definition.normalization == 'N3D'
    assert definition.channel_convention == 'ACN'
    assert definition.condon_shortley is False


@pytest.mark.parametrize("phase_convention", [True, False])
def test_setter_phase_convention(phase_convention):
    """Test setting Condon-Shortley phase convention with boolean values."""
    sph_harm = SphericalHarmonicDefinition()
    sph_harm.condon_shortley = phase_convention
    assert sph_harm.condon_shortley == phase_convention


def test_init_phase_convention_auto():
    """Test initialization with auto Condon-Shortley phase convention."""
    sph_harm = SphericalHarmonicDefinition(
        basis_type="complex",
        condon_shortley="auto")
    assert sph_harm.condon_shortley is True

    sph_harm = SphericalHarmonicDefinition(
        basis_type="real",
        condon_shortley="auto")
    assert sph_harm.condon_shortley is False

def test_setter_phase_convention_auto():
    """Test setting Condon-Shortley phase convention to auto."""
    sph_harm = SphericalHarmonicDefinition()
    sph_harm.basis_type = "complex"
    sph_harm.condon_shortley = "auto"
    assert sph_harm.condon_shortley is True

    sph_harm = SphericalHarmonicDefinition()
    sph_harm.basis_type = "real"
    sph_harm.condon_shortley = "auto"
    assert sph_harm.condon_shortley is False


def test_setter_phase_convention_invalid():
    """Test error handling for invalid Condon-Shortley phase values."""
    sph_harm = SphericalHarmonicDefinition()
    with pytest.raises(ValueError, match='must be a bool or the string'):
        sph_harm.condon_shortley = 123  # Invalid type

    with pytest.raises(ValueError, match='must be a bool or the string'):
        sph_harm.condon_shortley = "invalid"  # Invalid string


@pytest.mark.parametrize("channel_convention", ["ACN", "FuMa"])
def test_setter_channel_convention_definition(channel_convention):
    """Test setting channel convention for spherical harmonic definition."""
    sph_harm = SphericalHarmonicDefinition()
    sph_harm.channel_convention = channel_convention
    assert sph_harm.channel_convention == channel_convention


@pytest.mark.parametrize("channel_convention", ["ACN", "FuMa"])
def test_init_channel_convention_definition(channel_convention):
    """Test initialization with different channel conventions."""
    sph_harm = SphericalHarmonicDefinition(
        channel_convention=channel_convention)
    assert sph_harm.channel_convention == channel_convention


def test_setter_channel_convention_fuma_error():
    """Test error when setting FUMA channel convention with n_max > 3."""
    sph_harm = SphericalHarmonicDefinition(n_max=4)

    message = 'n_max > 3 is not allowed with'

    with pytest.raises(ValueError, match=message):
        sph_harm.channel_convention = "FuMa"

    with pytest.raises(ValueError, match=message):
        SphericalHarmonicDefinition(
            n_max=4,
            channel_convention="FuMa",
        )


def test_setter_channel_convention_definition_invalid():
    """Test error handling for invalid channel convention values."""
    sph_harm = SphericalHarmonicDefinition()
    with pytest.raises(ValueError, match='Invalid channel convention'):
        sph_harm.channel_convention = "invalid"  # Invalid value


@pytest.mark.parametrize("normalization", ["N3D", "SN3D", "maxN", "NM", "SNM"])
def test_setter_normalization_definition(normalization):
    """Test setting different normalization conventions."""
    sph_harm = SphericalHarmonicDefinition()
    sph_harm.normalization = normalization
    assert sph_harm.normalization == normalization


@pytest.mark.parametrize("normalization", ["N3D", "SN3D", "maxN", "NM", "SNM"])
def test_init_normalization_definition(normalization):
    """Test initialization with different normalization conventions."""
    sph_harm = SphericalHarmonicDefinition(
        normalization=normalization)
    assert sph_harm.normalization == normalization


def test_setter_normalization_definition_invalid():
    """Test error handling for invalid normalization values."""
    sph_harm = SphericalHarmonicDefinition()
    with pytest.raises(ValueError, match='Invalid normalization'):
        sph_harm.normalization = "invalid"  # Invalid value


def test_setter_normalization():
    """Test error when setting maxN normalization with n_max > 3."""
    sph_harm = SphericalHarmonicDefinition(n_max=4)

    with pytest.raises(ValueError, match='n_max > 3 is not allowed with'):
        sph_harm.normalization = "maxN"  # Invalid with n_max > 3


def test_sh_definition_setter_n_max():
    """Test setting n_max property and error handling for invalid values."""
    sph_harm = SphericalHarmonicDefinition(n_max=2)
    sph_harm.n_max = 3
    assert sph_harm.n_max == 3

    with pytest.raises(ValueError, match='n_max must be a positive integer'):
        sph_harm.n_max = -1  # Invalid value


@pytest.mark.parametrize(
        ('norm', 'convention'), [('maxN', 'ACN'), ('N3D', 'FuMa')])
def test_sh_definition_setter_n_max_invalid_combinations(norm, convention):
    """Test error when setting n_max > 3 with incompatible combinations."""
    sph_harm = SphericalHarmonicDefinition(n_max=2)

    sph_harm.channel_convention = convention
    sph_harm.normalization = norm
    with pytest.raises(ValueError, match='n_max > 3 is not allowed'):
        sph_harm.n_max = 4


def test_setter_basis_type():
    """Test setting basis type and error handling for invalid values."""
    sph_harm = SphericalHarmonicDefinition()
    sph_harm.basis_type = "complex"
    assert sph_harm.basis_type == "complex"

    sph_harm.basis_type = "real"
    assert sph_harm.basis_type == "real"

    with pytest.raises(ValueError, match='Invalid basis type'):
        sph_harm.basis_type = "invalid"  # Invalid value


@pytest.mark.parametrize(
        ("points", "method"), [
            (SamplingSphere([1, -1], [0, 0], [0, 0]), 'pseudo_inverse'),
            (pf.Coordinates(1, 0, 0), None),
        ],
    )
def test_sphharm_init(points, method):
    """Test default behaviour after initialization."""
    n_max = 0
    sph_harm = SphericalHarmonics(n_max=n_max, coordinates=points)
    assert sph_harm.n_max == n_max
    assert np.all(sph_harm.coordinates == points)
    assert sph_harm.inverse_method == method


def test_sphharm_init_invalid_coordinates():
    """Test error handling for invalid coordinate types."""
    with pytest.raises(TypeError,
                       match="coordinates must be a pyfar.Coordinates " \
                       "object or spharpy.SamplingSphere object"):
        SphericalHarmonics(n_max=2, coordinates=[0, 0, 1])

def test_sphharm_init_invalid_inverse_method(icosahedron_sampling):
    """Test error handling for invalid inverse_method values."""
    with pytest.raises(ValueError, match="Invalid inverse_method."):
        SphericalHarmonics(
            n_max=2, coordinates=icosahedron_sampling,
            inverse_method='invalid')

    coords = pf.Coordinates(1, 0, 0)
    with pytest.raises(ValueError, match="method can only be set if"):
        SphericalHarmonics(
            n_max=0, coordinates=coords,
            inverse_method='pseudo_inverse')


def test_sphharm_init_invalid_n_max(icosahedron_sampling):
    """Test error handling for invalid n_max values."""
    with pytest.raises(ValueError, match='n_max must be a positive integer'):
        SphericalHarmonics(n_max=-1, coordinates=icosahedron_sampling)

def test_sphharm_init_from_definition():
    """Test initialization using SphericalHarmonicDefinition."""

    n_max = 0
    coordinates = SamplingSphere(1, 0, 0)
    inverse_method = "auto"

    # generate from definition
    definition = SphericalHarmonicDefinition(n_max)
    sh_from_definition = SphericalHarmonics.from_definition(
        definition, coordinates, inverse_method)

    # generate manually
    sh_manually = SphericalHarmonics(
        n_max, coordinates, definition.basis_type, definition.normalization,
        definition.channel_convention, inverse_method,
        definition.condon_shortley)

    assert not deepdiff.DeepDiff(
        sh_from_definition.__dict__, sh_manually.__dict__)

def test_sphharm_init_from_definition_error():
    """Test error when passing wrong type to `from_definition` class method."""

    message = "definition must be a SphericalHarmonicDefinition"

    with pytest.raises(TypeError, match=message):
        SphericalHarmonics.from_definition(
            "definition", SamplingSphere(1, 0, 0))

def test_sphharm_compute_basis(icosahedron_sampling):
    """Test spherical harmonic basis computation."""
    sph_harm = SphericalHarmonics(n_max=2, coordinates=icosahedron_sampling)
    assert sph_harm.basis is not None

def test_sphharm_compute_basis_gradient(icosahedron_sampling):
    """Test spherical harmonic basis gradient computation."""
    sph_harm = SphericalHarmonics(n_max=2, coordinates=icosahedron_sampling)
    assert sph_harm.basis_gradient_theta is not None
    assert sph_harm.basis_gradient_phi is not None

def test_sphharm_compute_inverse_quad(icosahedron_sampling):
    """Test spherical harmonic inverse computation using quadrature method."""
    sh = SphericalHarmonics(
        1, coordinates=icosahedron_sampling, inverse_method='quadrature')
    assert sh.basis_inv is not None

def test_sphharm_compute_inverse_pseudo_inv(icosahedron_sampling):
    """Test spherical harmonic inverse using pseudo-inverse method."""
    sh = SphericalHarmonics(2, coordinates=icosahedron_sampling,
                            inverse_method='pseudo_inverse')
    assert sh.basis_inv is not None

def test_compute_basis_caching(icosahedron_sampling):
    """Test that basis computation results are cached and invalidated."""
    n_max = 1
    sh = SphericalHarmonics(
        n_max, icosahedron_sampling,
        basis_type='real',
        normalization='N3D',
        channel_convention='ACN',
        condon_shortley=False,
    )

    # Call the method once and store the result
    last_result = sh.basis

    sh.n_max = n_max  # Setting to the same value should not reset cache
    # Call the method again and check that the result is the same (cache hit)
    np.testing.assert_array_equal(sh.basis, last_result)

    # Change a property that affects the output of _compute_basis()
    sh.n_max = 3
    new_result = sh.basis

    # Call the method again and check that the result is different
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        new_result, last_result)
    last_result = new_result

    # check if changing the normalization triggers a cache reset
    sh.normalization = 'SN3D'
    new_result = sh.basis
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        new_result, last_result)
    last_result = new_result

    # check if changing the channel ordering triggers a cache reset
    sh.channel_convention = 'FuMa'
    new_result = sh.basis
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        new_result, last_result)
    last_result = new_result

    # check if changing the basis type triggers a cache reset
    sh.basis_type = 'complex'
    new_result = sh.basis
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        new_result, last_result)
    last_result = new_result

    # check if changing the Condon-Shortley phase convention triggers a reset
    sh.condon_shortley = True
    new_result = sh.basis
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        new_result, last_result)


def test_setter_inverse_method(icosahedron_sampling):
    """Test setting inverse method and error handling for invalid values."""
    sph_harm = SphericalHarmonics(n_max=2, coordinates=icosahedron_sampling)
    sph_harm.inverse_method = "quadrature"
    assert sph_harm.inverse_method == "quadrature"

    with pytest.raises(
            ValueError,
            match=("Invalid inverse_method. Allowed: 'pseudo_inverse', "
                   "'quadrature', or 'auto'.")):
        sph_harm.inverse_method = "invalid"  # Invalid value



import numpy as np
import pytest

import spharpy.samplings as samplings
from spharpy import SamplingSphere
from pyfar import Coordinates
import numpy.testing as npt
from spharpy.spherical import (
    spherical_harmonic_basis_real, spherical_harmonic_basis)


@pytest.mark.parametrize("flatten_output", [True, False])
def test_equidistant_cuboid_sampling_int(flatten_output):
    n_points = 3
    coords = samplings.equidistant_cuboid(
        n_points, flatten_output=flatten_output)
    data = np.linspace(-1, 1, 3)
    x, y, z = np.meshgrid(data, data, data, indexing='ij')
    if flatten_output:
        x = x.flatten()
        y = y.flatten()
        z = z.flatten()
    np.testing.assert_allclose(x, coords.x)
    np.testing.assert_allclose(y, coords.y)
    np.testing.assert_allclose(z, coords.z)
    assert type(coords) is Coordinates
    assert coords.csize == 3**3
    if flatten_output:
        assert coords.cshape == (3*3*3,)
    else:
        assert coords.cshape == (3, 3, 3)


def test_equidistant_cuboid_sampling_tuple():
    c = samplings.equidistant_cuboid((2, 3, 4), flatten_output=False)
    assert c.csize == 2*3*4
    assert c.cshape == (2, 3, 4)
    npt.assert_allclose(c.x[0], c.x[0, 0, 0])
    npt.assert_allclose(c.y[:, 0], c.y[0, 0, 0])
    npt.assert_allclose(c.z[:, :, 0], c.z[0, 0, 0])


def test_equidistant_cuboid_sampling_tuple_flatten():
    c = samplings.equidistant_cuboid((2, 3, 4), flatten_output=True)
    assert c.csize == 2*3*4
    assert c.cshape == (2*3*4,)


def test_equidistant_cuboid_sampling_invalid():
    with pytest.raises(ValueError, match='flatten_output must be a boolean.'):
        samplings.equidistant_cuboid(3, flatten_output='bla')
    with pytest.raises(ValueError, match='The number of points needs to be'):
        samplings.equidistant_cuboid(-3)
    with pytest.raises(ValueError, match='The number of points needs to be'):
        samplings.equidistant_cuboid((3, -3, 3))
    with pytest.raises(ValueError, match='The number of points needs to be'):
        samplings.equidistant_cuboid((3, 3, 3.2))
    with pytest.raises(ValueError, match='The number of points needs to be'):
        samplings.equidistant_cuboid((3, 3, -3))


def test_t_design_const_e(download_sampling):
    n_max = 2
    degree = 2 * n_max
    download_sampling('t-design', np.arange(1, 11))
    coords = samplings.t_design(
        n_max, criterion='const_energy')
    assert type(coords) is SamplingSphere
    assert coords.n_max == 2
    assert coords.csize == int(np.ceil((degree + 1)**2 / 2) + 1)


def test_t_design_const_angle(download_sampling):
    n_max = 2
    download_sampling('t-design', np.arange(1, 11))
    coords = samplings.t_design(
        n_max, criterion='const_angular_spread')
    assert type(coords) is SamplingSphere
    assert coords.csize == 18


def test_t_design_invalid(download_sampling):
    download_sampling('t-design', np.arange(1, 11))
    with pytest.raises(ValueError, match='Invalid design'):
        samplings.t_design(2, criterion='bla')


def test_t_design_limits_const_energy(download_sampling):
    download_sampling('t-design', [179, 180])
    samplings.t_design(90, criterion='const_energy')
    with pytest.raises(
            ValueError,
            match='n_max must be between 1 and 90 for const_energy'):
        samplings.t_design(91, criterion='const_energy')
    with pytest.raises(
            ValueError,
            match='n_max must be between 1 and 90 for const_energy'):
        samplings.t_design(0, criterion='const_energy')


def test_t_design_limits_const_angular_spread(download_sampling):
    download_sampling('t-design', [179, 180])
    samplings.t_design(89, criterion='const_angular_spread')
    with pytest.raises(
            ValueError,
            match='n_max must be between 1 and 89 for const_angular_spread'):
        samplings.t_design(90, criterion='const_angular_spread')
    with pytest.raises(
            ValueError,
            match='n_max must be between 1 and 89 for const_angular_spread'):
        samplings.t_design(0, criterion='const_angular_spread')


def test_t_design_n_max_error(download_sampling):
    download_sampling('t-design', np.arange(1, 11))
    samplings.t_design(89)
    with pytest.raises(
            ValueError,
            match='n_max must be an integer'):
        samplings.t_design(1.5)


def test_sph_t_design(download_sampling):
    # load test data
    download_sampling('t-design', np.arange(1, 11))

    # test with n_max
    c = samplings.t_design(1)
    isinstance(c, SamplingSphere)
    assert type(c) is SamplingSphere
    assert c.csize == 6

    # test default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test user radius
    c = samplings.t_design(2, radius=1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)

    # test loading degree order > 9
    c = samplings.t_design(10)

    # test quadrature
    assert not c.quadrature

    # test exceptions
    with pytest.raises(ValueError, match='Invalid design criterion'):
        samplings.t_design(2, criterion='const_thread')
    with pytest.raises(ValueError, match='radius must be a positive number'):
        samplings.t_design(2, radius=-1)
    with pytest.raises(ValueError, match='radius must be a positive number'):
        samplings.t_design(2, radius='test')


def test_dodecahedron():
    sampling = samplings.dodecahedron()
    assert type(sampling) is SamplingSphere


def test_sph_dodecahedron():
    # test with default radius
    c = samplings.dodecahedron()
    assert type(c) is SamplingSphere
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test with user radius
    c = samplings.dodecahedron(1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)

    # test quadrature
    assert not c.quadrature


def test_icosahedron():
    sampling = samplings.icosahedron()
    assert type(sampling) is SamplingSphere

    # test quadrature
    assert not sampling.quadrature


def test_sph_icosahedron():
    # test with default radius
    c = samplings.icosahedron()
    assert type(c) is SamplingSphere
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test with user radius
    c = samplings.icosahedron(1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)


def test_equiangular():
    # test without parameters
    with pytest.raises(ValueError, match='Either the n_points or n_max needs'):
        samplings.equiangular()

    # test with single number of points
    c = samplings.equiangular(n_points=5)
    assert type(c) is SamplingSphere
    assert c.csize == 5**2

    # test with tuple
    c = samplings.equiangular(n_points=(3, 5))
    assert c.csize == 3*5

    # test with spherical harmonic order
    c = samplings.equiangular(5)
    assert c.csize == 4 * (5 + 1)**2
    npt.assert_allclose(np.sum(c.weights), 4*np.pi)

    # test default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test user radius
    c = samplings.equiangular(5, radius=1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)


@pytest.mark.parametrize("n_points", np.arange(2, 40, 2))
def test_equiangular_weights_n_points_even(n_points):
    sampling = samplings.equiangular(n_points=n_points)
    npt.assert_almost_equal(np.sum(sampling.weights), 4*np.pi)
    assert sampling.cshape == sampling.weights.shape
    assert sampling.cshape == n_points*n_points
    assert sampling.quadrature


@pytest.mark.parametrize("n_points", np.arange(1, 40, 2))
def test_equiangular_weights_n_points_odd(n_points):
    sampling = samplings.equiangular(n_points=n_points)
    assert sampling.weights is None
    assert sampling.cshape == n_points*n_points
    assert not sampling.quadrature


@pytest.mark.parametrize(
    "n_points",
    [(5, 5), (4, 5), (4, 6)],
)
def test_equiangular_weights_n_points_tuple_invalid(n_points):
    sampling = samplings.equiangular(n_points=n_points)
    assert sampling.weights is None
    assert sampling.quadrature is False


def test_equiangular_weights_n_points_tuple_valid():
    n_points = (4, 4)
    sampling = samplings.equiangular(n_points=n_points)
    npt.assert_almost_equal(np.sum(sampling.weights), 4*np.pi)


@pytest.mark.parametrize("n_max", np.arange(1, 15))
def test_equiangular_weights_n_max(n_max):
    sampling = samplings.equiangular(n_max=n_max)
    npt.assert_almost_equal(np.sum(sampling.weights), 4*np.pi)
    assert sampling.cshape == sampling.weights.shape
    assert sampling.cshape == 4*(n_max+1)**2
    assert type(sampling) is SamplingSphere


@pytest.mark.parametrize(
    'basis_func', [spherical_harmonic_basis, spherical_harmonic_basis_real])
def test_equiangular_orthogonality(basis_func):
    n_max = 4
    sampling = samplings.equiangular(n_max=n_max)

    Y = basis_func(n_max, sampling)
    npt.assert_allclose(
        Y.conj().T @ np.diag(sampling.weights) @ Y,
        np.eye((n_max+1)**2),
        atol=1e-6, rtol=1e-6)


@pytest.mark.parametrize("n_max", np.arange(1, 15))
def test_gaussian_weights_n_max(n_max):
    sampling = samplings.gaussian(n_max=n_max)
    npt.assert_almost_equal(np.sum(sampling.weights), 4*np.pi)
    assert sampling.cshape == sampling.weights.shape
    assert sampling.cshape == 2*(n_max+1)*(n_max+1)
    assert type(sampling) is SamplingSphere


@pytest.mark.parametrize(
    'basis_func', [spherical_harmonic_basis, spherical_harmonic_basis_real])
def test_gaussian_orthogonality(basis_func):
    n_max = 4
    sampling = samplings.gaussian(n_max=n_max)

    Y = basis_func(n_max, sampling)
    npt.assert_allclose(
        Y.conj().T @ np.diag(sampling.weights) @ Y,
        np.eye((n_max+1)**2),
        atol=1e-6, rtol=1e-6)


def test_gaussian_quadrature():
    n_max = 3
    sampling = samplings.gaussian(n_max=n_max)

    assert sampling.quadrature


def test_gaussian():

    # test with spherical harmonic order
    c = samplings.gaussian(n_max=5)
    assert c.csize == 2 * (5 + 1)**2
    npt.assert_allclose(np.sum(c.weights), 4*np.pi)

    # test default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test user radius
    c = samplings.gaussian(5, radius=1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)

    # test quadrature
    npt.assert_allclose(np.sum(c.weights), 4 * np.pi)
    assert c.quadrature


def test_em32():
    sampling = samplings.eigenmike_em32()
    assert type(sampling) is SamplingSphere


def test_icosahedron_ke4():
    sampling = samplings.icosahedron_ke4()
    assert type(sampling) is SamplingSphere

    # test quadrature
    assert not sampling.quadrature


def test_equalarea():
    sampling = samplings.equal_area(2)
    assert type(sampling) is SamplingSphere

    # test quadrature
    assert not sampling.quadrature


def test_spiral_points():
    sampling = samplings.spiral_points(2)
    assert type(sampling) is SamplingSphere

    # test quadrature
    assert not sampling.quadrature


def test_equal_angle():
    # test with tuple
    c = samplings.equal_angle((10, 20))
    assert type(c) is SamplingSphere
    # test with number
    c = samplings.equal_angle(10)
    # test default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)
    # test user radius
    c = samplings.equal_angle(10, 1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)

    # test assertions
    with pytest.raises(ValueError,
                       match='delta_phi must be an integer divisor'):
        c = samplings.equal_angle((11, 20))
    with pytest.raises(ValueError,
                       match='delta_theta must be an integer divisor'):
        c = samplings.equal_angle((20, 11))

    # test quadrature
    assert not c.quadrature


def test_great_circle():
    # test with default values
    c = samplings.great_circle()
    assert type(c) is SamplingSphere
    # check default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test if azimuth matching angles work
    c = samplings.great_circle(0, 4, match=90)
    azimuth = c.azimuth * 180 / np.pi
    for deg in [0, 90, 180, 270]:
        assert deg in azimuth

    # test user radius
    c = samplings.great_circle(radius=1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)

    # test fractional azimuth resolution
    c = samplings.great_circle(60, 4,  azimuth_res=.1, match=90)
    npt.assert_allclose(c.azimuth[1] * 180 / np.pi, 7.5, atol=1e-15)

    # test quadrature
    assert not c.quadrature

    # test assertion: 1 / azimuth_res is not an integer
    with pytest.raises(AssertionError):
        samplings.great_circle(azimuth_res=.6)
    # test assertion: 360 / match is not an integer
    with pytest.raises(AssertionError):
        samplings.great_circle(match=270)
    # test assertion: match / azimuth_res is not an integer
    with pytest.raises(AssertionError):
        samplings.great_circle(azimuth_res=.5, match=11.25)


def test_lebedev(capfd):
    # test without parameters
    assert samplings.lebedev() is None
    # test command line output
    out, _ = capfd.readouterr()
    assert 'Possible input values' in out
    assert 'SH order 1, number of points 6' in out

    # test with spherical harmonic order
    c = samplings.lebedev(3)
    assert c.csize == 26
    assert type(c) is SamplingSphere

    # test default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)

    # test user radius
    c = samplings.lebedev(1, radius=1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)


@pytest.mark.parametrize("degree", np.array([
    6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302], dtype=int))
def test_lebedev_orthogonality(degree):
    """
    Test orthogonality of the transform.

    This was done after discovering https://github.com/pyfar/spharpy/issues/276
    to make sure that the sampling can be used for SH transforms using the
    pseudo inverse.

    The test for all degrees takes too long and was only done once. If required
    replace the above with

    @pytest.mark.parametrize("degree", np.array([
    6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350, 434,
    590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890, 4334,
    4802, 5294, 5810], dtype=int))
    """

    n_max = int(np.sqrt(degree / 1.3) - 1)
    sampling = samplings.lebedev(n_max)
    Y = spherical_harmonic_basis_real(n_max, sampling)
    Y_inverse = np.linalg.pinv(Y)

    # if orthogonal Y_inverse @ Y must be the identity matrix
    npt.assert_allclose(Y_inverse @ Y, np.eye((n_max + 1)**2), atol=1e-14)


@pytest.mark.parametrize(
        "n_max",
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
            11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29,
        ])
def test_fliege_for_each_order(n_max):
    c = samplings.fliege(n_max)
    assert type(c) is SamplingSphere
    assert c.csize == (n_max+1)**2
    npt.assert_allclose(c.radius, 1, atol=1e-15)


@pytest.mark.parametrize("n_max", [
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21,
    22, 23, 24, 25, 26, 27, 28, 29])
def test_fliege_orthogonality(n_max):
    """
    Test orthogonality of the transform.

    This was done after discovering https://github.com/pyfar/spharpy/issues/202
    to make sure that the sampling can be used for SH transforms using the
    pseudo inverse.
    """

    sampling = samplings.fliege(n_max)
    Y = spherical_harmonic_basis_real(n_max, sampling)
    Y_inverse = np.linalg.pinv(Y)

    # if orthogonal Y_inverse @ Y must be the identity matrix
    npt.assert_allclose(Y_inverse @ Y, np.eye((n_max + 1)**2), atol=1e-12)


def test_fliege_radius():
    # test user radius
    c = samplings.fliege(1, radius=1.5)
    npt.assert_allclose(c.radius, 1.5, atol=1e-15)


def test_fliege_errors():
    # test exceptions
    with pytest.raises(ValueError, match='n_max must be an integer.'):
        samplings.fliege(0.5)
    with pytest.raises(ValueError, match='n_max must be between 1 and 29'):
        samplings.fliege(0)
    with pytest.raises(ValueError, match='radius must be a positive number.'):
        samplings.fliege(1, radius=-1)
    with pytest.raises(ValueError, match='n_max must be between 1 and 29'):
        samplings.fliege(30)


def test_em64():
    sampling = samplings.eigenmike_em64()
    assert type(sampling) is SamplingSphere

    npt.assert_allclose(
        np.sum(sampling.weights), 4*np.pi, atol=1e-6, rtol=1e-6)


def test_hyperinterpolation_default(download_sampling):
    n_max = 1
    # download just required sampling for testing
    download_sampling('hyperinterpolation', [n_max])

    c = samplings.hyperinterpolation(1)

    # test sampling properties
    assert type(c) is SamplingSphere
    assert c.n_max == n_max
    assert c.csize == (n_max+1)**2
    assert c.radius.size == (n_max+1)**2

    # test default radius
    npt.assert_allclose(c.radius, 1, atol=1e-15)
    npt.assert_allclose(np.sum(c.weights), 4 * np.pi)


@pytest.mark.parametrize("radius", [1, 5])
def test_hyperinterpolation_radius(download_sampling, radius):
    download_sampling('hyperinterpolation', [1])
    sampling = samplings.hyperinterpolation(1, radius=radius)
    assert type(sampling) is SamplingSphere
    npt.assert_allclose(sampling.radius, radius, atol=1e-15)


@pytest.mark.parametrize("n_max", [-1, 'one'])
def test_hyperinterpolation_errors_n_max(n_max):
    with pytest.raises(
            ValueError, match='n_max must be an integer between 1 and 200'):
        samplings.hyperinterpolation(n_max)


@pytest.mark.parametrize("radius", [-1, 'one'])
def test_hyperinterpolation_errors_radius(radius):
    with pytest.raises(
            ValueError, match='radius must be a single positive value'):
        samplings.hyperinterpolation(1, radius)


def test_sph_gaussian_higher_order():
    s = samplings.gaussian(n_max=121)
    assert s.csize == (2 * (121 + 1)**2)
    npt.assert_allclose(np.sum(s.weights), 4*np.pi)



"""
Tests for spherical harmonic basis and related functions.
"""
import spharpy.spherical as sh
import numpy as np
import numpy.testing as npt
import pytest
from pyfar import Coordinates
from spharpy import SamplingSphere


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
@pytest.mark.parametrize("normalization", ['N3D', 'NM', 'maxN', 'SN3D', 'SNM'])
@pytest.mark.parametrize("channel_convention", ['ACN', 'FuMa'])
@pytest.mark.parametrize("condon_shortley", [True, False, 'auto'])
def test_spherical_harmonic(make_coordinates, coordinate_object,
                            normalization, channel_convention,
                            condon_shortley):
    n_max = 1
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    rad = np.ones(3, dtype=float)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, rad)

    phase_conv_id = 'None' if not condon_shortley else 'Condon-Shortley'

    norm_id = normalization
    if normalization == 'NM' or normalization == 'N3D':
        norm_id = 'n3d'
    if normalization == 'SNM' or normalization == 'SN3D':
        norm_id = 'sn3d'

    Y = np.genfromtxt(f'./tests/data/Y_cmplx_{phase_conv_id}_'
                      f'{norm_id}_{channel_convention.lower()}.csv',
                      dtype=complex,
                      delimiter=',')
    if normalization in ('NM', 'SNM'):
        Y *= np.sqrt(4 * np.pi)

    basis = sh.spherical_harmonic_basis(n_max, coords,
                                        normalization=normalization,
                                        channel_convention=channel_convention,
                                        condon_shortley=condon_shortley)

    np.testing.assert_allclose(Y, basis, atol=1e-13)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
@pytest.mark.parametrize("condon_shortley", [True, False, 'auto'])
@pytest.mark.parametrize("channel_convention", ['ACN', 'FuMa'])
@pytest.mark.parametrize("normalization", ['N3D', 'SN3D'])
def test_spherical_harmonics_real(make_coordinates, coordinate_object,
                                  normalization, channel_convention,
                                  condon_shortley):
    n_max = 1
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    rad = np.ones(3, dtype=float)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, rad)

    phase_conv_id = 'Condon-Shortley' if condon_shortley is True else 'None'

    norm_id = normalization
    if normalization == 'NM' or normalization == 'N3D':
        norm_id = 'n3d'
    if normalization == 'SNM' or normalization == 'SN3D':
        norm_id = 'sn3d'

    Y = np.genfromtxt(f'./tests/data/Y_real_{phase_conv_id}_'
                      f'{norm_id}_{channel_convention.lower()}.csv',
                      dtype=float,
                      delimiter=',')

    if normalization in ('NM', 'SNM'):
        Y *= np.sqrt(4 * np.pi)

    basis = sh.spherical_harmonic_basis_real(n_max, coords,
                                             normalization,
                                             channel_convention,
                                             condon_shortley=condon_shortley)
    np.testing.assert_allclose(basis, Y, atol=1e-13)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonics_invalid_nmax(make_coordinates, coordinate_object):
    n_max = 4
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    rad = np.ones(3, dtype=float)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, rad)

    with pytest.raises(ValueError,
                       match='MaxN normalization is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis(n_max, coords,
                                    normalization='maxN')
    with pytest.raises(ValueError,
                       match='MaxN normalization is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis_real(n_max, coords,
                                         normalization='maxN')


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonics_invalid_fuma(make_coordinates, coordinate_object):
    n_max = 4
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    rad = np.ones(3, dtype=float)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, rad)

    with pytest.raises(ValueError,
                       match='FuMa channel convention is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis(n_max, coords,
                                    channel_convention='FuMa')
    with pytest.raises(ValueError,
                       match='FuMa channel convention is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis_real(n_max, coords,
                                         channel_convention='FuMa')


@pytest.mark.parametrize("coordinate_object", [SamplingSphere, Coordinates])
def test_spherical_harmonics_invalid_condon_shortley(
        make_coordinates, coordinate_object,
    ):
    n_max = 4
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    rad = np.ones(3, dtype=float)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, rad)

    with pytest.raises(ValueError,
                       match="Condon_shortley has to be a bool, or 'auto'."):
        sh.spherical_harmonic_basis(n_max, coords,
                                    condon_shortley='xx')
    with pytest.raises(ValueError,
                       match="Condon_shortley has to be a bool, or 'auto'."):
        sh.spherical_harmonic_basis_real(n_max, coords,
                                         condon_shortley='xx')


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonic_default_n10(make_coordinates, coordinate_object):
    """Test the default parameters of SH basis function generator. This
    simultaneously tests if the methods still match the implementation
    up to spharpy 0.6.2.
    """
    n_max = 10
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))

    Y = np.genfromtxt(
        './tests/data/sh_basis_cplx_n10.csv',
        delimiter=',',
        dtype=complex)

    basis = sh.spherical_harmonic_basis(n_max, coords)

    np.testing.assert_allclose(Y, basis, atol=1e-13)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonics_real_n10_default(
        make_coordinates, coordinate_object,
    ):
    """Test the default parameters of SH basis function generator. This
    simultaneously tests if the methods still match the implementation
    up to spharpy 0.6.2.
    """
    n_max = 10
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2], dtype=float)
    phi = np.array([0, np.pi/2, 0, np.pi/4], dtype=float)
    rad = np.ones(4)

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, rad)

    Y = np.genfromtxt('./tests/data/sh_basis_real.csv',
                      dtype=float,
                      delimiter=',')
    basis = sh.spherical_harmonic_basis_real(n_max, coords)
    np.testing.assert_allclose(basis, Y, atol=1e-13)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_orthogonality(make_coordinates, coordinate_object):
    """
    Check if the orthonormality condition of the spherical harmonics is
    fulfilled.
    """
    n_max = 82
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2], dtype=float)
    phi = np.array([0, np.pi/2, 0, np.pi/4], dtype=float)
    n_points = phi.size
    n_points = len(theta)
    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))
    basis = sh.spherical_harmonic_basis(n_max, coords)

    inner = (basis @ np.conjugate(basis.T))
    fact = 4 * np.pi / (n_max + 1) ** 2
    orth = np.diagonal(fact * inner)
    np.testing.assert_allclose(
        orth, np.ones(n_points, dtype=complex), rtol=1e-10)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_orthogonality_real(make_coordinates, coordinate_object):
    """
    Check if the orthonormality condition of the reavl valued spherical
    harmonics is fulfilled.
    """
    n_max = 82
    theta = np.array([np.pi / 2, np.pi / 2, 0, np.pi / 2], dtype='double')
    phi = np.array([0, np.pi / 2, 0, np.pi / 4], dtype='double')
    n_points = phi.size

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))

    basis = sh.spherical_harmonic_basis_real(n_max, coords)

    inner = (basis @ np.conjugate(basis.T))
    fact = 4 * np.pi / (n_max + 1) ** 2
    orth = np.diagonal(fact * inner)
    np.testing.assert_allclose(orth, np.ones(n_points), rtol=1e-10, atol=1e-10)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonic_basis_gradient(
        make_coordinates, coordinate_object,
    ):
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))

    grad_ele, grad_azi = \
        sh.spherical_harmonic_basis_gradient(n_max, coords)

    desire_ele = np.genfromtxt(
        './tests/data/Y_grad_ele.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_ele, desire_ele, rtol=1e-10, atol=1e-10)

    desire_azi = np.genfromtxt(
        './tests/data/Y_grad_azi.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_azi, desire_azi, rtol=1e-10, atol=1e-10)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonic_basis_gradient_real(
        make_coordinates, coordinate_object,
    ):
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))

    grad_ele, grad_azi = \
        sh.spherical_harmonic_basis_gradient_real(n_max, coords)

    desire_ele = np.genfromtxt(
        './tests/data/Y_grad_real_ele.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_ele, desire_ele, rtol=1e-10, atol=1e-10)

    desire_azi = np.genfromtxt(
        './tests/data/Y_grad_real_azi.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_azi, desire_azi, rtol=1e-10, atol=1e-10)


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonics_gradient_invalid_nmax(
        make_coordinates, coordinate_object,
    ):
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))

    with pytest.raises(ValueError,
                       match='MaxN normalization is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis_gradient(n_max, coords,
                                             normalization='maxN')
    with pytest.raises(ValueError,
                       match='MaxN normalization is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis_gradient_real(n_max, coords,
                                                  normalization='maxN')


@pytest.mark.parametrize("coordinate_object", [Coordinates, SamplingSphere])
def test_spherical_harmonics_gradient_invalid_fuma(
        make_coordinates, coordinate_object,
    ):
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    coords = make_coordinates.create_coordinates(
        coordinate_object, phi, theta, np.ones_like(theta))

    with pytest.raises(ValueError,
                       match='FuMa channel convention is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis_gradient(n_max, coords,
                                             channel_convention='FuMa')
    with pytest.raises(ValueError,
                       match='FuMa channel convention is only'
                             ' supported up to 3rd order.'):
        sh.spherical_harmonic_basis_gradient_real(n_max, coords,
                                                  channel_convention='FuMa')



import numpy as np
import spharpy.samplings as samplings
from spharpy.special import spherical_bessel_zeros
from scipy.special import spherical_jn


def test_sph_bessel_zeros():
    roots = spherical_bessel_zeros(3, 3)
    jn_zeros = np.ones(roots.shape)
    for n in range(4):
        jn_zeros[n, :] = spherical_jn(n, roots[n, :])
    zeros = np.zeros((4, 3), dtype=float)
    np.testing.assert_allclose(jn_zeros, zeros, atol=1e-12)


def test_interior_points_chardon():
    kr_max = 4
    int_points = samplings.interior_stabilization_points(kr_max)

    truth = np.array([[0, 0, 0]], dtype=float)

    np.testing.assert_allclose(int_points.cartesian, truth, atol=1e-7)



"""Spherical samplings."""

from .samplings import (
    equidistant_cuboid,
    hyperinterpolation,
    t_design,
    dodecahedron,
    icosahedron,
    icosahedron_ke4,
    equiangular,
    gaussian,
    eigenmike_em32,
    equal_area,
    spiral_points,
    equal_angle,
    great_circle,
    lebedev,
    fliege,
    eigenmike_em64,
)

from .helpers import (
    spherical_voronoi,
    calculate_sampling_weights,
)

from .interior import interior_stabilization_points


__all__ = [
    'equidistant_cuboid',
    'hyperinterpolation',
    't_design',
    'dodecahedron',
    'icosahedron',
    'icosahedron_ke4',
    'equiangular',
    'gaussian',
    'eigenmike_em32',
    'equal_area',
    'spiral_points',
    'spherical_voronoi',
    'calculate_sampling_weights',
    'interior_stabilization_points',
    'equal_angle',
    'great_circle',
    'lebedev',
    'fliege',
    'eigenmike_em64',
]



"""Wave field representations in the spatial domain."""
from .spatial import greens_function_plane_wave, greens_function_point_source


__all__ = 'greens_function_plane_wave', 'greens_function_point_source'



import numpy as np
import numpy.testing as npt
import pytest
import pyfar as pf

import spharpy


def test_dolph_cheby_mainlobe():
    N = 7
    theta0 = np.pi/6
    d_nm = spharpy.beamforming.dolph_chebyshev_weights(
        N, theta0, design_criterion='mainlobe')

    truth = np.loadtxt('tests/data/dolph_cheby_mainlobe.csv', delimiter=',')
    npt.assert_allclose(d_nm, truth)


def test_dolph_cheby_sidelobe():
    N = 7
    R_dB = 50
    R = 10**(R_dB/20)
    d_nm = spharpy.beamforming.dolph_chebyshev_weights(
        N, R, design_criterion='sidelobe')

    truth = np.loadtxt('tests/data/dolph_cheby_sidelobe.csv', delimiter=',')
    npt.assert_allclose(d_nm, truth)


def test_re_max():
    N = 7
    g_nm = spharpy.beamforming.rE_max_weights(N, normalize=False)

    truth = np.loadtxt('tests/data/re_max_weights.csv', delimiter=',')
    npt.assert_allclose(g_nm, truth)

    g_nm_norm = spharpy.beamforming.rE_max_weights(N, normalize=True)

    Y = spharpy.spherical.spherical_harmonic_basis_real(
        N, pf.Coordinates(1, 0, 0))

    npt.assert_allclose(Y @ np.diag(g_nm_norm) @ Y.T, 1)


def test_max_front_back():
    N = 7
    f_nm_norm = spharpy.beamforming.maximum_front_back_ratio_weights(
        N, normalize=True)

    Y = spharpy.spherical.spherical_harmonic_basis_real(
        N, pf.Coordinates(1, 0, 0))

    npt.assert_allclose(Y @ np.diag(f_nm_norm) @ Y.T, 1)

    with pytest.raises(RuntimeError, match='did not converge'):
        spharpy.beamforming.maximum_front_back_ratio_weights(30)


def test_normalize_weights():
    n_max = 7
    normalized_weights = spharpy.beamforming.normalize_beamforming_weights(
        np.ones(n_max+1), n_max)

    # pwd weights are const 4*pi/(n_max+1)**2
    expected = np.ones(n_max+1)*4*np.pi/(n_max+1)**2

    npt.assert_allclose(expected, normalized_weights)
